"""
This module implements a stripes plot function that can create annual and monthly
stripes plots from given climate and weather data. There is a matplotlib and a 
plotly version. The streamlit app uses the plotly version by default.
The plotly plot is interactive and available in a German and English version, the
matplotlib version is static and only available in English.


"""
from typing import Tuple
import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
import seaborn as sns
import plotly.graph_objects as go


def yrplot_data(sr, ref=[1980, 2000], ag='sum', qa=0.95, cmxeq=True) -> Tuple[pd.DataFrame, float, float]:
    """
    Prepare the given data for monthly and annual stripes to be plotted in a heatmap.
    The data is grouped and the limits for plotting are returned, based on the quantile given.

    """
    # plot of heatmap with monthyl and annual stripes
    yrs = sr.index.year.unique()
    dummy = np.zeros((len(yrs), 14)) * np.nan
    dummy = pd.DataFrame(dummy)
    dummy.index = yrs
    dummy.columns = ['J', 'F', 'M', 'A', 'M', 'J', 'J', 'A', 'S', 'O', 'N', 'D', '', 'Year']

    for i in sr.index:
        dummy.iloc[i.year - yrs[0], i.month - 1] = sr.loc[i]
    for i in yrs:
        if ag == 'sum':
            dummy.iloc[i - yrs[0], 13] = sr.loc[sr.index.year == i].sum()
            cmx = 'coolwarm_r'
        elif ag == 'min':
            dummy.iloc[i - yrs[0], 13] = sr.loc[sr.index.year == i].min()
        elif ag == 'max':
            dummy.iloc[i - yrs[0], 13] = sr.loc[sr.index.year == i].max()
        else:  # ag == 'mean'
            dummy.iloc[i - yrs[0], 13] = sr.loc[sr.index.year == i].mean()

    if ref == None:
        pass
    else:
        refx = dummy.loc[ref[0]:ref[1]].mean(axis=0)
        dummy = dummy - refx

    if cmxeq:
        vxU = dummy.abs().quantile(qa).quantile(qa)
        vxL = -1. * vxU
    else:
        vxU = dummy.quantile(qa).quantile(qa)
        vxL = dummy.quantile(1. - qa).quantile(1. - qa)

    if ag == 'sum':
        dummy.iloc[:, 13] = dummy.iloc[:, 13] / 12

    return dummy, vxU, vxL


def _yrplot_matplotlib(sr, ref=[1980, 2000], ag='sum', qa=0.95, cbar_title='Temperature anomaly (K)', cmx='coolwarm', cmxeq=True, li=False):
    # prepare the data
    dummy, vxU, vxL = yrplot_data(sr, ref=ref, ag=ag, qa=qa, cmxeq=cmxeq)
    yrs = dummy.index.values
    # build the figure
    fig = plt.figure(figsize=(8,len(dummy)/15.))
    ax = sns.heatmap(dummy, cmap=cmx, vmin=vxL, vmax=vxU, cbar_kws={'label': cbar_title})

    if ref == None:
        pass
    else:
        ax.add_patch(
            plt.Rectangle((0, ref[0] - yrs[0]), 12, ref[1] - ref[0], fill=True, edgecolor='red', facecolor='gray', lw=2,
                          alpha=0.3, clip_on=False))
        ax.add_patch(
            plt.Rectangle((13, ref[0] - yrs[0]), 1, ref[1] - ref[0], fill=True, edgecolor='red', facecolor='gray', lw=2,
                          alpha=0.3, clip_on=False))
        ax.annotate('Reference period', (0.5, ref[1] - yrs[0] - 2), color='white', weight='bold', ha='left',
                    va='bottom', alpha=0.6)

    if type(li) == int:
        ax.axhline(li - yrs[0], color='k', ls='--', lw=1, alpha=0.5)
        ax.annotate(' >> observed', (12.5, li - yrs[0] - 0.5), color='k', ha='center', va='bottom', alpha=0.6,
                    rotation=90.)
        ax.annotate('modelled << ', (12.5, li - yrs[0] + 0.5), color='k', ha='center', va='top', alpha=0.6,
                    rotation=90.)
        # ax.add_patch(plt.Rectangle((0, li-yrs[0]), 12, 0, fill=False, edgecolor='k', ls='--', lw=1, alpha=0.5, clip_on=False))

    ax.set_ylabel('Year')
    ax.set_xlabel('Month          ')
    
    return fig


def _yrplot_plotly(sr, ref=[1980, 2000], ag='sum', qa=0.95, cbar_title='Temperature anomaly (K)', cmx='coolwarm', cmxeq=True, li=False, lang='en'):
    # prepare the data
    dummy, vxU, vxL = yrplot_data(sr, ref=ref, ag=ag, qa=qa, cmxeq=cmxeq)
    
    # use full names
    if lang == 'de':
        dummy.columns = ['Janua', 'Februar', 'März', 'April', 'Mai', 'Juni', 'Juli', 'August', 'September', 'Oktober', 'November', 'Dezember', '', 'Jahr']
        hv_tmpl = """<b>Jahr:</b> %{y}<br><b>Monat:</b> %{x}<br><b>Abweichung:</b> %{z:.1f} K"""
        ref_txt = f'<b>Reference Period</b><br>{ref[0]} - {ref[1]}'
    else:
        dummy.columns = ['January', 'February', 'March', 'April', 'May', 'June', 'July', 'August', 'September', 'October', 'November', 'December', '', 'Year']
        hv_tmpl = """<b>Year:</b> %{y}<br><b>Month:</b> %{x}<br><b>Deviation:</b> %{z:.1f} K"""
        ref_txt = f'<b>Reference Period</b><br>{ref[0]} - {ref[1]}'
    
    fig = go.Figure()
    
    if cmx == 'coolwarm':
        lo = min(0, abs(vxL / dummy.min().min()))
        up = min(1, vxU / dummy.max().max())
        cmx = [
            [lo, '#8FBDD3'],
            [(up + lo) / 2 , '#E4D1B9'],
            [up, '#A97155'],
        ]
    # add the trace
    fig.add_trace(go.Heatmap(
        z=dummy,
        x=dummy.columns,
        y=dummy.index,
        colorscale=cmx,
        hovertemplate=hv_tmpl,
        showlegend=False,
        name=cbar_title
    ))

    # add reference window
    fig.add_shape(type='rect', x0=-0.5, y0=ref[0], x1=11.5, y1=ref[1], xref='x', fillcolor='rgba(255, 0,0,0.3)', line=dict(width=3, color='red'))
    fig.add_shape(type='rect', x0=12.5, y0=ref[0], x1=13.5, y1=ref[1], xref='x', fillcolor='rgba(255, 0,0,0.3)', line=dict(width=3, color='red'))
    fig.add_annotation(x=12, y=np.mean(ref), showarrow=False, text=ref_txt, textangle=-90, font=dict(size=16, color='red'))

    # add colorbar title
    fig.add_annotation(x=1.1, align='right', valign='top', xref='paper', yref='paper', xanchor='right', yanchor='middle', textangle=-90, text=cbar_title, showarrow=False, font=dict(size=16, color='black'))
    
    # update the layout
    fig.update_layout(
        template='plotly_white',
        height=min(len(dummy) * 15, 650),
        margin=dict(t=25, b=100, l=80, r=160, autoexpand=False),
        xaxis=dict(tickangle=45, title='Monat' if lang == 'de' else 'Month'),
        yaxis=dict(title='Jahr' if lang == 'de' else 'Year'),
    )

    # return
    return fig


def yrplot_hm(sr, ref=[1980, 2000], ag='sum', qa=0.95, cbar_title='Temperature anomaly (K)', cmx='coolwarm', cmxeq=True, li=False, lang='en', backend='plotly'):
    """
    Yearly stripes plot.
    Creates a monthly resloved stripes plot for each year in the data.
    The data is plotted as a heatmap showing the Temperature anomaly to the passed reference period
    based on the adjustable quartile value.

    Parameters
    ----------
    sr : pandas.DataFrame
        Input data originating from one ground station
    ref : Tuple[int, int]
        Reference period to calculate the anomaly.
    ag : str
        Reduction function to use. Can be used to calculate
        the temperature anomaly in 'sum', 'min', 'max' or 'mean'
    qa : float
        Quartile to use. Has to be between 0 and 1.
    cbar_title : str
        Optional title for the colorbar
    cmx : str
        Colorbar scale. Can be any string accepted by the plotting backend.
    cmxeq : bool
        If True (default), the colorbar range is calculated for 
        the full dataset.
    lang : str
        Can be either ``'en'`` or ``'de'`. The language used for axis labels.
    backend : str
        Can be either ``matplotlib`` or ``plotly``.
        The plotting backend to use.

    """
    if backend.lower() == 'matplotlib':
        return _yrplot_matplotlib(sr, ref=ref, ag=ag, qa=qa, cbar_title=cbar_title, cmx=cmx, cmxeq=cmxeq, li=li)
    else:
        return _yrplot_plotly(sr, ref=ref, ag=ag, qa=qa, cbar_title=cbar_title, cmx=cmx, cmxeq=cmxeq, li=li, lang=lang)



# SIMPLEcrop in python 3
# RUINS Project (cc) c.jackisch@tu-braunschweig.de

# This is a python implementation based on Zhao et al. 2019 and other crop, soilwater and evapotranspiration models
# The intention is a to port the SIMPLE crop model to a fast and modular modelling system
# Zhao, C. et al. (2019), A SIMPLE crop model, EUROPEAN JOURNAL OF AGRONOMY, 104, 97–106, doi:10.1016/j.eja.2019.01.009.

import numpy as np
import pandas as pd
import numba

def dTT(T,Tbase):
    #Temperature with reference to a base for degree-day method after Ritchie et al. 1985 (CERES-Wheat)
    return np.fmax(T-Tbase,0.)

def dBM(fSolar, SRAD, F_CO2, F_Temp, F_Water, F_Heat, RUE):
    #change in biomass similar to Williams et al. 1989 (EPIC)
    dBiomass = 10. * RUE * fSolar * SRAD * F_CO2 * F_Temp * np.fmin(F_Water, F_Heat)
    return dBiomass

def fsolar(TT,F_Water,fSolar_water,Tsum,I50maxW,I50maxH,fsolmax=0.95):
    #Beer-Lambert’s law of light attenuation
    I50B = TT.copy() * 0.
    fSolar = TT.copy() * 0.
    for i in np.arange(len(TT))[1:]:
        fSol1 = np.min([1.,fsolmax/(1+np.exp((TT.iloc[i]-I50B.iloc[i-1])*-0.01))])
        fSol2 = np.min([1.,fsolmax / (1 + np.exp((TT.iloc[i]-(Tsum - I50B.iloc[i-1])) * 0.01))])
        fSolar.iloc[i] = np.min([fSol1,fSol2])*np.min([fSolar_water.iloc[i-1],1.])

        I50B.iloc[i] = np.max([np.max([I50B.iloc[i - 1] + I50maxW*(1.-F_Water.iloc[i-1]),I50B.iloc[i - 1] + I50maxH*(1.-F_Water.iloc[i-1])]),0.])
        if TT.iloc[i] > Tsum:
            maturityday = TT.iloc[i]
            #print('crop maturity')
            break
        if (fSolar.iloc[i]<fSolar.iloc[i-1]) & (fSolar.iloc[i]<=0.005):
            maturityday = TT.index[i]
            #print('crop maturity due to senecence')
            break
    return fSolar,I50B


def T_resp(tmean,Tbase,Topt):
    Tr = np.fmax((tmean-Tbase)/(Topt-Tbase),0.)
    Tr[tmean>=Topt] = 1.
    return Tr

def heat_resp(tmax,maxT,extremeT):
    #heat stress factor based on  Asseng et al., 2011 (APSIM-Nwheat)
    F_Heat = np.fmax(1. - (tmax - maxT) / (extremeT - maxT), 0.)
    F_Heat[tmax > extremeT] = 0.
    F_Heat[tmax <= maxT] = 1.
    return F_Heat

def water_resp(arid,s_water):
    #water balance approach
    return np.fmax(0.,1.-s_water*arid)

def arid(ETo,PAW):
    #water balance aridity index after Woli et al. (2012)
    return 1. - np.min([ETo,0.096*PAW])/ETo

def CO2_resp(CO2, CO2_RUE):
    if (CO2 >= 700.):
        return (1 + CO2_RUE * 350 / 100)
    else:
        return np.max([(CO2_RUE * CO2 * 0.01 + 1 - 0.01 * 350. * CO2_RUE), 1.])

def I50Bp1(I50B,Imaxwater,arid,swater):
    fwater = water_resp(arid, swater)
    return I50B + Imaxwater * (1. - fwater)


### EVAPOTRANSPIRATION ###
# the ETo function can refer to different models.
# I provide the daily ET Szilagyi Jozsa model as first entry

@numba.jit
def Te_opt(T_e,gammax, vabar):
    maxdeltaT_e = 1.
    maxit = 9999999
    itc = 0
    while maxdeltaT_e < 0.01:
        v_e = 0.6108 * np.exp(17.27 * T_e / (T_e + 237.3))  # saturated vapour pressure at T_e (S2.5)
        T_enew = gammax * (v_e - vabar)  # rearranged from S8.8
        deltaT_e = T_enew - T_e
        T_e = T_enew
        maxdeltaT_e = np.abs(np.max(deltaT_e))
        if itc > maxit:
            break
        itc += 1
    return T_e



def ET_SzilagyiJozsa(data, Elev, lat, windfunction_ver='1948', alpha=0.23, zerocorr=True):
    # Taken from R package Evapotranspiration >> Danlu Guo <danlu.guo@adelaide.edu.au>
    # Daily Actual Evapotranspiration after Szilagyi, J. 2007, doi:10.1029/2006GL028708
    # data is assumed to be a pandas data frame with at least:
    # T or Tmin/max - daily temperature in degree Celcius,
    # RH or RHmin/max - daily  relative humidity in percentage,
    # u2 - daily wind speed in meters per second
    # Rs - daily solar radiation in Megajoule per square meter.
    # Result in [mm/day] for EToSJ, EToPM, and EToPT reference ET

    # update alphaPT according to Szilagyi and Jozsa (2008)
    alphaPT = 1.31
    lat_rad = lat * np.pi / 180.
    alphaPT = 1.31
    sigma = 4.903e-09
    Gsc = 0.082
    lambdax = 2.45

    # add julian Days
    J = data.index.dayofyear

    # Calculating mean temperature
    if ('Ta' in data.columns):
        Ta = data['Ta']
    elif ('T' in data.columns):
        Ta = data['T']
    elif ('tas' in data.columns):
        Ta = data['tas']
    else:
        if ('Tmax' in data.columns):
            Ta = (data['Tmax'] + data['Tmin']) / 2  
        else:
            Ta = (data['tasmax'] + data['tasmin']) / 2  
            # Equation S2.1 in Tom McMahon's HESS 2013 paper, which in turn was based on Equation 9 in Allen et al, 1998.


    # Saturated vapour pressure
    if ('Tmax' in data.columns):
        vs_Tmax = 0.6108 * np.exp(17.27 * data['Tmax'] / (data['Tmax'] + 237.3))  # Equation S2.5
        vs_Tmin = 0.6108 * np.exp(17.27 * data['Tmin'] / (data['Tmin'] + 237.3))  # Equation S2.5
        vas = (vs_Tmax + vs_Tmin) / 2.  # Equation S2.6
    else:
        vs_Tmax = 0.6108 * np.exp(17.27 * data['tasmax'] / (data['tasmax'] + 237.3))  # Equation S2.5
        vs_Tmin = 0.6108 * np.exp(17.27 * data['tasmin'] / (data['tasmin'] + 237.3))  # Equation S2.5
        vas = (vs_Tmax + vs_Tmin) / 2.  # Equation S2.6
    # Vapour pressure
    if 'RHmax' in data.columns:
        vabar = (vs_Tmin * data['RHmax'] / 100. + vs_Tmax * data['RHmin'] / 100.) / 2.  # Equation S2.7
        # mean relative humidity
        RHmean = (data['RHmax'] + data['RHmin']) / 2

    else:
        if ('RH' in data.columns):
            vabar = (vs_Tmin + vs_Tmax) / 2 * data['RH'] / 100.
            RHmean = data['RH']
        elif ('huss' in data.columns):
            vabar = (vs_Tmin + vs_Tmax) / 2 * data['huss'] / 100.
            RHmean = data['huss']
        elif ('hurs' in data.columns):
            vabar = (vs_Tmin + vs_Tmax) / 2 * data['hurs'] / 100.
            RHmean = data['hurs']

        # Calculations from data and constants for Penman


    P = 101.3 * ((293. - 0.0065 * Elev) / 293.) ** 5.26  # atmospheric pressure (S2.10)
    delta = 4098 * (0.6108 * np.exp((17.27 * Ta) / (Ta + 237.3))) / (
                (Ta + 237.3) ** 2)  # slope of vapour pressure curve (S2.4)
    gamma = 0.00163 * P / lambdax  # psychrometric constant (S2.9)
    d_r2 = 1 + 0.033 * np.cos(2 * np.pi / 365 * J)  # dr is the inverse relative distance Earth-Sun (S3.6)
    delta2 = 0.409 * np.sin(2 * np.pi / 365 * J - 1.39)  # solar dedication (S3.7)
    w_s = np.arccos(-1. * np.tan(lat_rad) * np.tan(delta2))  # sunset hour angle (S3.8)
    N = 24 / np.pi * w_s  # calculating daily values
    R_a = (1440 / np.pi) * d_r2 * Gsc * (
                w_s * np.sin(lat_rad) * np.sin(delta2) + np.cos(lat_rad) * np.cos(delta2) * np.sin(
            w_s))  # extraterristrial radiation (S3.5)
    R_so = (0.75 + (2 * 10 ** - 5) * Elev) * R_a  # clear sky radiation (S3.4)

    if 'Rs' in data.columns:
        R_s = data['Rs']
    elif 'rsds' in data.columns:
        R_s = data['rsds']
    else:
        print('Radiation data missing')
        return

    if ('Tmax' in data.columns):
        R_nl = sigma * (0.34 - 0.14 * np.sqrt(vabar)) * ((data['Tmax'] + 273.2) ** 4 + (data['Tmin'] + 273.2) ** 4) / 2 * (
                1.35 * R_s / R_so - 0.35)  # estimated net outgoing longwave radiation (S3.3)
    else:
        R_nl = sigma * (0.34 - 0.14 * np.sqrt(vabar)) * ((data['tasmax'] + 273.2) ** 4 + (data['tasmin'] + 273.2) ** 4) / 2 * (
                1.35 * R_s / R_so - 0.35)  # estimated net outgoing longwave radiation (S3.3)
    # For vegetated surface
    R_nsg = (1 - alpha) * R_s  # net incoming shortwave radiation (S3.2)
    R_ng = R_nsg - R_nl  # net radiation (S3.1)

    if 'u2' in data.columns:
        u2 = data['u2']
        if windfunction_ver == "1948":
            f_u = 2.626 + 1.381 * u2  # wind function Penman 1948 (S4.11)
        elif windfunction_ver == "1956":
            f_u = 1.313 + 1.381 * u2  # wind function Penman 1956 (S4.3)

        Ea = f_u * (vas - vabar)  # (S4.2)

        Epenman_Daily = delta / (delta + gamma) * (R_ng / lambdax) + gamma / (
                    delta + gamma) * Ea  # Penman open-water evaporation (S4.1)
    
    elif 'sfcWind' in data.columns:
        u2 = data['sfcWind']
        if windfunction_ver == "1948":
            f_u = 2.626 + 1.381 * u2  # wind function Penman 1948 (S4.11)
        elif windfunction_ver == "1956":
            f_u = 1.313 + 1.381 * u2  # wind function Penman 1956 (S4.3)

        Ea = f_u * (vas - vabar)  # (S4.2)

        Epenman_Daily = delta / (delta + gamma) * (R_ng / lambdax) + gamma / (
                    delta + gamma) * Ea  # Penman open-water evaporation (S4.1)
    
    else:

        Epenman_Daily = 0.047 * R_s * np.sqrt(Ta + 9.5) - 2.4 * (R_s / R_a) ** 2 + 0.09 * (Ta + 20) * (
                    1 - RHmean / 100)  # Penman open-water evaporation without wind data by Valiantzas (2006) (S4.12)

    # Iteration for equilibrium temperature T_e
    T_e = Ta
    gammax = Ta - 1 / gamma * (1 - R_ng / (lambdax * Epenman_Daily))

    T_e = Te_opt(T_e.values, gammax.values, vabar.values)
    T_e = pd.Series(T_e, index=Ta.index)

    deltaT_e = 4098 * (0.6108 * np.exp((17.27 * T_e) / (T_e + 237.3))) / (
                (T_e + 237.3) ** 2)  # slope of vapour pressure curve (S2.4)
    E_PT_T_e = alphaPT * (deltaT_e / (deltaT_e + gamma) * R_ng / lambdax)  # Priestley-Taylor evapotranspiration at T_e
    E_SJ_Act_Daily = 2 * E_PT_T_e - Epenman_Daily  # actual evapotranspiration by Szilagyi and Jozsa (2008) (S8.7)

    if zerocorr:
        ET_Daily = np.fmax(E_SJ_Act_Daily, 0.)
    else:
        ET_Daily = E_SJ_Act_Daily

    return ET_Daily,Epenman_Daily,E_PT_T_e


### INFILTRATION ###
# the function for available soil water content can be set to different solvers here
# if there is no time series provided (which might be the common case)
# for now, I stick to the original ARIDITY estimate


def fARID(data, para, ETfu='SJ'):
    '''
    Simple soil water update based on original SIMPLE code
    :param AWC: available water capacity (Vol/Vol)
    :param DDC: deep drainage coefficient (-)
    :param RCN: runoff curve number (-)
    :param RZD: rootzone depth (mm)
    :param WUC: water uptake coefficient (-)
    :switch ETfu: function for ET can be SJ, PM or PT
    '''

    #caluclate ETo if not existing
    if 'EToSJ' in data.columns:
        if ETfu == 'PT':
            ETO = data.EToPT
        elif ETfu == 'PM':
            ETO = data.EToPM2
        else:
            ETO = data.EToSJ
    elif 'Rs' in data.columns:
        EToSJ,EToPM,EToPT = ET_SzilagyiJozsa(data, para['Elev'], para['lat'])
        if ETfu == 'PT':
            ETO = EToPT
        elif ETfu == 'PM':
            ETO = EToPM
        else:
            ETO = EToSJ
    else:
        print('NO RADIATION DATA IN DATAFRAME! CANNOT RUN CROP MODEL.')

    #calculate potential direct runoff
    RO = (data['Prec'] - 0.2 * (25400 / para['RCN'] - 254)) ** 2 / (data['Prec'] + 0.8 * (25400 / para['RCN'] - 254))
    RO[data['Prec'] <= 0.2 * (25400 / para['RCN'] - 254)] = 0.
    CWBD = data['Prec'] - RO

    WBD = CWBD.cumsum() + (para['RZD'] * para['AWC'])
    WBD[WBD / para['RZD'] > para['AWC']] -= para['RZD'] * para['DDC'] * (
                WBD[WBD / para['RZD'] > para['AWC']] / para['RZD'] - para['AWC'])

    WAT = CWBD.copy() * 0.
    ARID = WAT.copy() * 0.
    WAT.iloc[0] = para['RZD'] * para['AWC']
    for i in np.arange(len(CWBD))[1:]:
        WAT.iloc[i] = WAT.iloc[i - 1] + CWBD.iloc[i]
        if (WAT.iloc[i] / para['RZD'] > para['AWC']):
            WAT.iloc[i] -= np.min([para['RZD'] * para['DDC'] * (WAT.iloc[i] / para['RZD'] - para['AWC']), WAT.iloc[i]])
        TR = np.min([para['WUC'] * para['RZD'] * WAT.iloc[i] / para['RZD'], ETO.iloc[i]])
        WAT.iloc[i] -= TR
        if (TR > 0.) & (ETO.iloc[i] > 0.):
            ARID.iloc[i] = 1. - TR / ETO.iloc[i]
        else:
            ARID.iloc[i] = 1. - TR

    return [ARID, ETO]


def SIMPLE(data, sowing, stopday, para, CO2):
    '''
    SIMPLE crop model after Zhao et al. 2019
    :para[' data']: data frame of input weather data
    :para[' sowing']: time stamp of sowing
    :para[' stopday']: time stamp of end of simulation
    :para[' para'][' dictionary'] of para['eters']
    :return: time series of crop development
    '''

    ARID, ETO = fARID(data.loc[sowing:stopday], para, ETfu='SJ')
    F_Heat = heat_resp(data.loc[sowing:stopday, 'Tmax'], para['maxT'], para['extremeT'])
    F_Temp = T_resp(data.loc[sowing:stopday, 'T'], para['Tbase'], para['Topt'])
    F_CO2 = CO2_resp(CO2, para['CO2_RUE'])

    TT = dTT(data.loc[sowing:stopday, 'T'], para['Tbase']).cumsum() + para['InitialTT']

    F_Water = water_resp(ARID, para['s_water'])
    fSolar_water = F_Water
    fSolar_water[fSolar_water > 0.1] = 1.
    fSolar_water[
        fSolar_water <= 0.1] += 0.9  # radiation interception is affected when the drought stress becomes severe enough after Steduto et al. (2009) (AquaCrop model)

    fSolar, I50B = fsolar(TT, F_Water, fSolar_water, para['Tsum'], para['I50maxW'], para['I50maxH'], fsolmax=0.95)
    BM = (10. * para['RUE'] * fSolar * data.loc[sowing:stopday, 'Rs'] * F_CO2 * F_Temp * np.fmin(F_Water,
                                                                                               F_Heat)).cumsum() + para['IniBio']  # change in biomass similar to Williams et al. 1989 (EPIC)
    Yield = BM * para['HIp']

    res = pd.concat([Yield, ARID, ETO, fSolar, I50B], axis=1)
    res.columns = ['Yield', 'ARID', 'ETO', 'fSolar', 'I50B']
    return res



import plotly.graph_objects as go
import pandas as pd


def sunburst(df: pd.DataFrame, maxdepth: int = 4, fig: go.Figure = None, width: int = 700, height: int = 700) -> go.Figure:
    """
    Create a sunburst plot of all climate models included in the current
    dataset used in RUINS. It will group them by GCM -> RCM -> RCP.
    The size of the sun can be restricted by the maxdepth argument.
    """
    # extract the data
    ids = df.ids.values
    labels = df.labels.values
    parents = df.parents.values
    values = df['values'].values
    customdata = list(zip(df.customdata1.values, df.customdata2.values))

    # handle figure
    if fig is None:
        fig = go.Figure()

    # build the sunburst
    fig.add_trace(go.Sunburst(
        ids=ids,
        labels=labels,
        parents=parents,
        values=values,
        branchvalues='total',
        insidetextorientation='radial',
        #hoverinfo='label+percent root+percent parent+value',
        maxdepth=maxdepth,
        customdata=customdata,
        hovertemplate='<b>Name</b>: %{label}<br><b>Total</b>: %{value}<br><br>%{customdata[0]}<br>%{customdata[1]}',
        name=''
    ))

    # figure config
    fig.update_layout(margin=dict(t=0, l=0, r=0, b=0),  width=width, height=height)
    
    return fig



from collections import defaultdict

import pandas as pd

from ruins.core import DataManager


def ordered_sunburst_data(dataManager: DataManager, order=['GCM', 'RCM', 'RCP']) -> pd.DataFrame:
    """Order the info about GCM / RCM /RCP usage amounts"""
    # TODO: this can be updated later
    NAME = 'RUINS'

    # get the data 
    climate_models = dataManager.read('climate')

    # get all combinations
    combinations = [climate_models[v].attrs for v in climate_models.data_vars]
    N = len(combinations)

    # create the aggreated data structure
    agg = defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: 0)))

    # fill the structure - this is the crucial part for order
    for c in combinations:
        agg[c[order[0]]][c[order[1]]][c[order[2]]] += 1

    # final lists for the sunburn plot
    ids = [NAME]
    labels = [NAME]
    parents = [""]
    values = [N]
    customdata = [(f'100% of {NAME}', '-', )]

    # Go for each order element and add as a element to the lists
    for lvl1 in agg.keys():
        ids.append(f'{NAME}.{lvl1}')
        labels.append(lvl1)
        parents.append(NAME)
        lvl1_val = sum([sum([r for r in lvl2.values()]) for lvl2 in agg[lvl1].values()])
        values.append(lvl1_val)
        customdata.append((f'{int(lvl1_val / N * 100)}% of {NAME}', '-', ))

        # level 2
        for lvl2 in agg[lvl1].keys():
            ids.append(f'{NAME}.{lvl1}.{lvl2}')
            labels.append(lvl2)
            parents.append(f'{NAME}.{lvl1}')
            lvl2_val = sum([r for r in agg[lvl1][lvl2].values()])
            values.append(lvl2_val)
            customdata.append((f'{int((lvl2_val / N * 100))}% of {NAME}', f'{int(lvl2_val / lvl1_val * 100)}% of {lvl1}', ))

            # level 3
            for lvl3 in agg[lvl1][lvl2].keys():
                ids.append(f'{NAME}.{lvl1}.{lvl2}.{lvl3}')
                labels.append(lvl3.upper())
                parents.append(f'{NAME}.{lvl1}.{lvl2}')
                val = agg[lvl1][lvl2][lvl3]
                values.append(val)
                customdata.append((f'{int(val / N * 100)}% of {NAME}', f'{int(val / lvl2_val * 100)}% of {lvl2}', ))
    
    # puh done
    return pd.DataFrame(dict(
        ids=ids,
        labels=labels,
        parents=parents,
        values=values,
        customdata1=[_[0] for _ in customdata],
        customdata2=[_[1] for _ in customdata]
    ))


"""
Distribution plots as used by the uncertainty application.
"""
from typing import List
from itertools import cycle

import numpy as np
from scipy import stats
import plotly.graph_objects as go
from plotly.express.colors import unlabel_rgb, label_rgb, n_colors, sequential


def distribution_plot(*events: List[dict], fig: go.Figure = None, **kwargs) -> go.Figure:
    """Concept plot for PDFs of different events.
    This plot illustrates the interdependece of Kngihtian uncertainty
    and risk for a set of events. At least one event has to be passed.
    
    Parameters
    ----------
    events : List
        Each event is represented by a dict, that has to contain the following keys:
        outcomes: List[Tuple[float, float]]
            The mu and std of each outcome for this event
        Additionally, the following keys are optional:
        dist : str
            The distribution of the event. Default is 'norm'.
        coloscale : str
            The colorscale used to color the events. Default is a 
            cycle through ['Greens', 'Reds', 'Blues']
        name : str
            Name for this event
    
    Returns
    -------
    fig : go.Figure
        Result plot
    """
    # create a figure
    if fig is None:
        fig = go.Figure()

    # get the x-axis
    x = np.linspace(0, 10, 200)

    # get the first event
    if len(events) == 0:
        raise ValueError('At least one event has specified')

    # default colorscale names
    COLORS = cycle(['Greens', 'Reds', 'Blues', 'Oranges', 'Greys'])

    for event in events:
        # get the distribution
        dist = getattr(stats, event.get('dist', 'norm'))

        # get the coloscale
        cscale = getattr(sequential, event.get('colorscale', next(COLORS)))
        cmap = n_colors(unlabel_rgb(cscale[-1]), unlabel_rgb(cscale[0]), len(event['outcomes']) + 2)

        for i, outcome in enumerate(event['outcomes']):
            mu, std = outcome
            y = dist.pdf(x, loc=mu, scale=std)
            y_sum = y.sum()
            y /= y_sum

            # add the traces
            fig.add_trace(
                go.Scatter(x=x, y=y * 100, 
                mode='lines', 
                line=dict(color=label_rgb(cmap[i])), 
                name=event.get('name', f'Outcome #{i + 1}'), 
                fill='tozerox'
                )
            )
            fig.add_trace(
                go.Scatter(x=[mu, mu], y=[0, dist.pdf(mu, loc=mu, scale=std) / y_sum * 100], 
                mode='lines',
                line=dict(color=label_rgb(cmap[i]), width=3, dash='dash'),
                name=f"Mean {event.get('name', f'Outcome #{i + 1}')}",
                )
            )
    # adjust figure
    fig.update_layout(
        template='plotly_white',
        legend=dict(orientation='h')
    )

    # return
    return fig



import numpy as np
import matplotlib.pyplot as plt


def kde(data, cmdata='none', split_ts=1, cplot=True, eq_period=True):
    # plot of kde with stripes
    from sklearn.neighbors import KernelDensity
    cxx = ['#E69F00', '#009E73', '#0072B2', '#D55E00', '#CC79A7']
    cxx2 = ['#8c6bb1', '#810f7c']

    data = data[~np.isnan(data)]
    x_d = np.linspace(np.min(data) * 0.9, np.max(data) * 1.1, len(data))

    # instantiate and fit the KDE model
    kde = KernelDensity(bandwidth=1.0, kernel='gaussian')
    kde.fit(data[:, None])

    # score_samples returns the log of the probability density
    logprob = kde.score_samples(x_d[:, None])

    if cplot & (split_ts == 1):
        plt.fill_between(x_d, np.exp(logprob), alpha=0.4, facecolor='grey')

    lp = np.exp(logprob)
    xd = x_d

    if type(cmdata) != str:
        fig, (ax, cax, cax2) = plt.subplots(ncols=3, figsize=(10.5, 2.3), gridspec_kw={"width_ratios": [1, 0.02, 0.02]})

        x_d2 = np.linspace(np.min(cmdata) * 0.9, np.max(cmdata) * 1.1, len(cmdata))
        # instantiate and fit second KDE model
        kde2 = KernelDensity(bandwidth=1.0, kernel='gaussian')
        kde2.fit(cmdata[:, None])

        # score_samples returns the log of the probability density
        logprob2 = kde2.score_samples(x_d2[:, None])

        if cplot & (split_ts == 1):
            ax.fill_between(x_d2, np.exp(logprob2), alpha=0.4, facecolor='grey')

        lp2 = np.exp(logprob2)
        xd2 = x_d2
    else:
        fig, (ax, cax) = plt.subplots(ncols=2, figsize=(10.5, 2.3), gridspec_kw={"width_ratios": [1, 0.06]})

    if split_ts > 1:
        if eq_period:
            spliti = [0, len(data) - 40, len(data) - 20, len(data)]
        else:
            spliti = np.linspace(0, len(data), split_ts + 1).astype(int)

        for i in np.arange(split_ts):
            datax = data.iloc[spliti[i]:spliti[i + 1]]
            x_d = np.linspace(np.min(datax) * 0.9, np.max(datax) * 1.1, len(datax))

            try:
                # instantiate and fit the KDE model
                kde = KernelDensity(bandwidth=1.0, kernel='gaussian')
                kde.fit(datax[:, None])

                # score_samples returns the log of the probability density
                logprob = kde.score_samples(x_d[:, None])

                if cplot:
                    ax.fill_between(x_d, np.exp(logprob), alpha=0.4, facecolor=cxx[i],
                                    label='-'.join([str(datax.index.year.min()), str(datax.index.year.max())]))
            except:
                pass

    if type(cmdata) != str:
        # add climate model data
        csplit_ts = [2040, 2080]
        for i in np.arange(2):
            datax = cmdata.loc[(cmdata.index.year >= csplit_ts[i]) & (cmdata.index.year < csplit_ts[i] + 20)]
            x_d = np.linspace(np.min(datax) * 0.9, np.max(datax) * 1.1, len(datax))

            try:
                # instantiate and fit the KDE model
                kde = KernelDensity(bandwidth=1.0, kernel='gaussian')
                kde.fit(datax[:, None])

                # score_samples returns the log of the probability density
                logprob = kde.score_samples(x_d[:, None])

                if cplot:
                    ax.fill_between(x_d, np.exp(logprob), alpha=0.4, facecolor=cxx2[i],
                                    label='-'.join([str(datax.index.year.min()), str(datax.index.year.max())]))

            except:
                pass

    ax.legend(loc=1, ncol=5)

    if cplot:
        # cmap = plt.cm.get_cmap('cividis_r')
        cmap = plt.cm.get_cmap('viridis_r')
        colors = plt.cm.cividis_r(np.linspace(0, 1, len(data)))
        colorsx = cmap(np.arange(cmap.N))

        for i in np.arange(len(data)):
            ax.plot([data.iloc[i], data.iloc[i]], [0, np.max(lp) * 0.9], c=colors[i])

        labcb = 'Year'
        if type(cmdata) != str:
            cmap2 = plt.cm.get_cmap('plasma')
            colors2 = plt.cm.plasma(np.linspace(0, 1, len(cmdata)))
            colorsx2 = cmap2(np.arange(cmap2.N))

            for i in np.arange(len(cmdata)):
                ax.plot([cmdata.iloc[i], cmdata.iloc[i]], [0, np.max(lp) * 0.9], c=colors2[i])

            cbar2 = plt.colorbar(plt.cm.ScalarMappable(cmap=cmap2), cax=cax2, label=labcb,
                                 ticks=[0, 0.2, 0.4, 0.6, 0.8, 1], fraction=0.0027, anchor=(1.0, 0.1))
            cbar2.ax.set_yticklabels(
                np.round(np.linspace(cmdata.index.year.min(), cmdata.index.year.max(), 6)).astype(int).astype(str))
            labcb = ''

        cbar = plt.colorbar(plt.cm.ScalarMappable(cmap=cmap), cax=cax, label=labcb, ticks=[0, 0.2, 0.4, 0.6, 0.8, 1])
        cbar.ax.set_yticklabels(
            np.round(np.linspace(data.index.year.min(), data.index.year.max(), 6)).astype(int).astype(
                str))  # vertically oriented colorbar

    ax.set_ylabel('Occurrence (KDE)')

    return fig, ax



import plotly.graph_objects as go 
from plotly.subplots import make_subplots
import pandas as pd


DIMENSIONS = {l: dict() for l in ['January', 'February', 'March', 'April', 'May', 'June', 'July', 'August', 'September', 'October', 'November', 'December']}


def climate_projection_parcoords(data: pd.DataFrame, fig: go.Figure = None, align_range: bool = True, colorscale = 'electric', row: int = 1, col: int = 1, lang='en'):
    """
    Parallel coordinates plot for climate projections.
    This plot uses each month in the year as a coordinate dimension. By sorting the
    dimensions into the correct order, the cycle of annual temperature aggregates is preserved,
    while the full dataset can easily be compared.

    Parameters
    ----------
    data : pandas.DataFrame
        Dataframe with the data to plot. The DataFrame has to be indexed by a
        Datetime Index and does accept more than one column (ground station, RCP scenario or grid cell).
    fig : plotly.graph_objects.Figure
        If not None, the given figure will be used to plot the data.
        Note, that subfigures need to use the ``'domain'`` type.
    align_range : bool
        If True (default) each dimension (aka month) will use the same value range, to
        focus the differences between the months. If False, the range will be adapted
        to span from min to max for each dimension, putting more focus on the differences
        between the years (decades).
    colorscale : str
        Name identifier of teh colorscale. See plotly to learn about available options.
    row : int
        If figure is not None, row and column can be used to plot into the 
        correct subplot.
    col: int
        If figure is not None, row and column can be used to plot into the
        correct subplot.
    lang : str
        Can either be ``'en'`` or ``'de'``. As of now, the language does not
        have any effect.

    """
    # create the dimensions dictionary
    dimensions = {k:v for k, v in DIMENSIONS.items()}

    # colormap container
    cmap = []
    vmin = data.min().min()
    vmax = data.max().max()

    # group by Month
    grp = data.groupby(data.index.map(lambda x: x.strftime('%B')))

    # create dimensions
    for label, df in grp:
        df = pd.melt(df, ignore_index=False).drop('variable', axis=1)
        dim = dict(label=label, values=df.values)
        if align_range:
            dim['range'] = (vmin, vmax)
        
        # append
        dimensions[label] = dim
        cmap.extend(df.index.map(lambda x: x.year).values.tolist())
    
    if fig is None:
        fig = make_subplots(1, 1, specs=[[{'type': 'domain'}]])

    # make plot
    fig.add_trace(go.Parcoords(
        line=dict(color=cmap, colorscale=colorscale, showscale=True),
        dimensions = [dict(label='Year', values=cmap)] + list(dimensions.values())
    ), row=row, col=col)

    return fig


from typing import Callable, Dict

DEFAULT = {}

def get_translator(lang: str, default: dict = dict(), **translations: Dict[str, str]) -> Callable[[str], str]:
    """Translator factory
    Returns a function that returns prepared translations from
    a languge dictionary for a passed key. The dictionaries have to be passed
    as kwargs, using the language as key. 
    """
    # check for the language
    if lang not in translations:
        if len(default.keys()) > 0:
            _T = default
        else:
            _T = DEFAULT
    else:
        _T = translations[lang]
    
    # define the helper
    def _get_translation(key: str):
        return _T.get(key, '')
    
    # return
    return _get_translation



"""
Data Select
===========

Alex will das dokumentieren....

"""
from typing import List, Union
import streamlit as st

from ruins.core import DataManager, Config
from ruins.plotting import plt_map


# I have no better idea for translation right now
_TRANSLATE_EN = dict(
    map_caption_long='''Map with available stations (<span style="color:blue">blue dots</span>) and selected reference station (<span style="color:magenta">magenta highlight</span>). The climate model grid is given in <span style="color:orange">orange</span> with the selected references as filled dots.''',
    map_caption_short='''Map with available stations (<span style="color:blue">blue dots</span>) and selected reference station (<span style="color:magenta">magenta highlight</span>).''',
    welcome='Welcome to Weather Explorer',
    rcp_title='Include Climate Projections?',
    rcp_description="""RCPs are scenarios about possible greenhouse gas concentrations by the year 2100. RCP2.6 is a world in which little further greenhouse gasses are emitted -- similar to the Paris climate agreement from 2015. RCP8.5 was intendent to explore a rather risky, more worst-case future with further increased emissions. RCP4.5 is one candidate of a more moderate greenhouse gas projection, which might be more likely to resemble a realistic situation. It is important to note that the very limited differentiation between RCP scenarios have been under debate for several years. One outcome is the definition of Shared Socioeconomic Pathways (SSPs) for which today, however, there are not very many model runs awailable. For more information, please check with the [Climatescenario Primer](https://climatescenarios.org/primer/), [CarbonBrief](https://www.carbonbrief.org/explainer-how-shared-socioeconomic-pathways-explore-future-climate-change) and this [NatureComment](https://www.nature.com/articles/d41586-020-00177-3)""",
    rcp_infotext="### Activate climate projections?\nYou can select one of the scenarios below, or select 'No' to skip this step, then no climate projections will be included. Each projection has a linked video from https://sos.noaa.gov/ to illustrate the changes due to the scenario."
)
_TRANSLATE_DE = dict(
    map_caption_long='''Karte mit verfügbaren Stationen (<span style="color:blue">blauen Punkte</span>) und ausgewählte Referenzstation (<span style="color:magenta">magenta hervorgehoben</span>). Das Modellgrid ist in <span style="color:orange">orange</span> mit den ausgewählten Referenzen als gefüllte Punkte dargestellt.''',
    map_caption_short='''Karte mit verfügbaren Stationen (<span style="color:blue">blauen Punkte</span>) und ausgewählte Referenzstation (<span style="color:magenta">magenta hervorgehoben</span>).''',
    welcome='Willkommen bei Weather Explorer',
    rcp_title='Klimaprojektionen hinzufügen?',
    rcp_description="""RCPs sind mögliche Szenarios der Treibhausgaskonentrationen in der Atmosphäre bis zum Jahr 2100. RCP2.6 orientiert sich am Pariser Klimaabkommen von 2015, demnach kaum mehr Treibhasugase ausgestoßen werden. RCP8.5 beschreibt einen eher risikofreudigen Umgang, bei dem Konzentrationen ähnlich einem Worst-Case Szenario weiter ansteigen. RCP4.5 is ein moderates Modell, das eher an der Realität orientiert ist.. Es ist wichtig zu beachten, dass die schwierige Differenzierbarkeit zwischen RCP Szenarios seit langem diskutiert wird. Ein Ausgang ist die Definition von Shared Socioeconomic Pathways (SSPs), für die heute jedoch nur wenige Modell-Runs verfügbar sind. Für weitere Informationen, siehe [Climatescenario Primer](https://climatescenarios.org/primer/), [CarbonBrief](https://www.carbonbrief.org/explainer-how-shared-socioeconomic-pathways-explore-future-climate-change) oder dieser [Nature Comment](https://www.nature.com/articles/d41586-020-00177-3) kontaktieren.""",
    rcp_inftext = "### Klimaprojektionen hinzufügen?\nSie können eines der Szenarios unten auswählen, oder 'Keine' auswählen, um diesen Schritt zu überspringen. Jedes Szenario hat ein verlinktes Video von https://sos.noaa.gov/ zur Darstellung der Änderungen durch das Szenario."
)

_VIDEO_SOURCES = {
    'default': dict(
        title='### {rcp}',
        description='Video Source: [NOAA Science On a Sphere](https://sos.noaa.gov/catalog/datasets/climate-model-temperature-change-rcp-{rcp}-2006-2100/)',
        video_url='https://sos.noaa.gov/videos/rcp_ga_{rcp}_400.mp4'
    ),
    'rcp26': dict(
        title='### RCP 2.6',
        description='Video Source: [NOAA Science On a Sphere](https://sos.noaa.gov/catalog/datasets/climate-model-temperature-change-rcp-26-2006-2100/)',
        video_url='https://sos.noaa.gov/videos/rcp_ga_26_400.mp4'
    ),
    'rcp45': dict(
        title='### RCP 4.5',
        description='Video Source: [NOAA Science On a Sphere](https://sos.noaa.gov/catalog/datasets/climate-model-temperature-change-rcp-45-2006-2100/)',
        video_url='https://sos.noaa.gov/videos/rcp_ga_45_400.mp4'
    ),
    'rcp85': dict(
        title='### RCP 8.5',
        description='Video Source: [NOAA Science On a Sphere](https://sos.noaa.gov/catalog/datasets/climate-model-temperature-change-rcp-85-2006-2100/)',
        video_url='https://sos.noaa.gov/videos/rcp_ga_85_400.mp4'
    ),
    'allRCP': dict(
        title='### all RCPs',
        description='All RCPs try to project CO2 concentrations, like shown in the video. Source: [NOAA Science On a Sphere](https://sos.noaa.gov/catalog/datasets/carbon-dioxide-concentration-geos-5-model/)',
        video_url='https://sos.noaa.gov/videos/geos_5_carbon_400_audio.mp4'
    ),
    'weather': dict(
        title='### Weather Data',
        description='Compare with weather observations. The video shows the current global precipitation. [SOURCE](https://sos.noaa.gov/catalog/datasets/clouds-with-precipitation-real-time/)',
        video_url='https://sos.noaa.gov/videos/clouds_precip_400.mp4'
    ),
}


def _map_wrapper(dataManager: DataManager, config: Config, expander_container = st.sidebar, add_caption: bool = True, **kwargs):
    # get the translator
    t = kwargs.get('translator', config.translator(en=_TRANSLATE_EN, de=_TRANSLATE_DE))

    # create correct map
    if config.get('include_climate', False):
        fig = plt_map(dataManager, sel=config.get('selected_station', 'all'), cm='CORDEX')
        text = t('map_caption_long')
    else:
        fig = plt_map(dataManager, sel=config.get('selected_station', 'all'))
        text = t('map_caption_short')

    # add
    expander_container.plotly_chart(fig, use_container_width=True)

    if add_caption:
        expander_container.markdown(text, unsafe_allow_html=True)


def selected_station_selector(dataManager: DataManager, config: Config, expander_container = st.sidebar, **kwargs):
    """
    Select a station.
    """
    container = st if 'container' not in kwargs else kwargs['container']
    t = kwargs.get('translator', config.translator(en=_TRANSLATE_EN, de=_TRANSLATE_DE))
    
    # get a station list
    if 'station_list' in kwargs:
        station_list = kwargs['station_list']
    else:
        weather = dataManager['weather'].read()
        station_list = list(weather.keys()) # TODO station names krummhoern, coast, inland, niedersachsen?    
    
    if config.has_key('selected_station'):
        _map_wrapper(dataManager, config, expander_container, add_caption=False)
        expander_container.selectbox('Station(s) to use', station_list, key='selected_station')
        return

    container.title(t('welcome'))
    
    # placehoder for map
    _map_wrapper(dataManager=dataManager, config=config, expander_container=container)
    
    # introduction
    container.markdown("HERE WE EXPLAIN EVERYTHING THE USER NEEDS TO KNOW ABOUT THE DATA")

    container.info("Right now, you can't select a station by clicking on it. Please use the select below")

    cap = 'Station(s) to use' if config.lang == 'en' else 'Messstation(en):'
    preselect = container.selectbox(cap, [' - select - '] + station_list)

    # check input
    if preselect ==  ' - select - ':
        st.stop()
    else:
        st.session_state.selected_station = preselect
        st.experimental_rerun()


def temporal_agg_selector(dataManager: DataManager, config: Config, expander_container = st.sidebar, **kwargs):
    """
    Select a temporal aggregation.
    """
    container = st if 'container' not in kwargs else kwargs['container']
    if config.has_key('temporal_agg'):
        expander_container.selectbox('Temporal aggregation', config.get('temporal_aggregations', ["Annual", "Monthly"]), key='temporal_agg')
        return

    container.title("Select a temporal aggregation")

    left, right = container.columns(2)

    left.info("I am a yearly aggregated preview")
    right.info("I am a monthly aggregated preview")

    use_annual = left.button('Jährliche Aggregierung' if config.lang == 'de' else 'Aggregate Annualy')
    use_monthly = right.button('Monatliche Aggregierung' if config.lang == 'de' else 'Aggregate Monthly')

    if use_annual:
        st.session_state.temporal_agg = "Annual"
    elif use_monthly:
        st.session_state.temporal_agg = "Monthly"
    else:
        st.stop()
    
    st.experimental_rerun()


def rcp_selector(dataManager: DataManager, config: Config, expander_container = st.sidebar, elements: Union[str, List[str]] = '__rpc__',  **kwargs):
    """
    Decide whether to include or exclude climate projects. 
    If climate projections should be included, the user can select one of the projections.
    """
    # RCP key to use - needed for multiple selects
    RCP_KEY = kwargs.get('RCP_KEY', 'current_rcp')

    # get the container
    container = st if 'container' not in kwargs else kwargs['container']
    t = kwargs.get('translator', config.translator(en=_TRANSLATE_EN, de=_TRANSLATE_DE))

    # check what to show as projections
    if elements == '__rpc__':
        # get all available climate projections
        projections = sorted(list(set([_.split('.')[-1] for _ in dataManager['cordex_coast'].read().keys()])))
    elif elements == '__all__':
        # get all available climate projections
        projections = sorted(list(set([_.split('.')[-1] for _ in dataManager['cordex_coast'].read().keys()])))
        projections = ['weather', 'allRCP'] + projections

    elif isinstance(elements, str):
        projections = [elements]
    else:
        projections = elements

    # check mode
    if config.has_key('include_climate') and kwargs.get('allow_skip', True):
        cap = 'Include climate projections' if config.lang == 'en' else 'Klimaprojektionen hinzufügen?'
        expander_container.checkbox(cap, key='include_climate')

    # add the selectbox
    if config.has_key(RCP_KEY):
        cap = 'Select RCP:' if config.lang == 'en' else 'RCP auswählen:'
        expander_container.selectbox(cap, projections, key=RCP_KEY, format_func=lambda x: x.upper())
        return
    
    # main content
    layout = kwargs.get('layout', 'columns')
    
    # title
    container.title(kwargs.get('title', t('rcp_title')))

    container.markdown(t('rcp_description'), unsafe_allow_html=True)

#    container.info("RPCs are cool, but make stuff a bit slower")
    container.markdown(t('rcp_infotext'))

    # check if the user can skip this step
    if kwargs.get('allow_skip', True):
        no = container.button('Ohne Klimaprojektionen fortfahren' if config.lang == 'de' else 'Continue without climate projections')

        # handle the no
        if no:
            st.session_state.include_climate = False
            st.experimental_rerun()
    
    # build the columns
    if layout == 'columns':
        cols = container.columns(len(projections))
    else:
        cols = [container.columns((1,2,2)) for _ in range(len(projections))]
    
    # container for click buttons
    use_rcp = []
    
    # go for it
    for col, rcp in zip(cols, projections):
        # get the layout and the data
        c = (col, col, col) if layout == 'columns' else col
        _dat = _VIDEO_SOURCES.get(rcp, _VIDEO_SOURCES['default'])
        
        print(rcp)
        # video
        c[0].video(_dat['video_url'].format(rcp=rcp[-2:]))
        
        # description
        c[1].markdown(_dat['title'].format(rcp=rcp.upper()))
        c[1].markdown(_dat['description'].format(rcp=rcp))
        
        # button
        use = c[2].button(f'Use {rcp.upper()}')
        use_rcp.append((use, rcp))
    
    # check if any button was used:
    for use, rcp in use_rcp:
        if use:
            st.session_state.include_climate = True
            st.session_state[RCP_KEY] = rcp
            st.experimental_rerun()
    
    # if we didn't restart, stop the application
    st.stop()


def data_select(dataManager: DataManager, config: Config, expander_container = st.sidebar, **kwargs):
    """
    Select a station, a temporal aggregation and if climate projections 
    should be included.
    If we are either not in story_mode or if all of the above are already 
    selected, new data can be selected in a selectbox in the sidebar 
    (compact_selector()).
    Otherwise data is selected step by step alongside with explanations.
    """
    # set default story mode 
    story_mode = config.get('story_mode', True)
    if 'story_mode' in kwargs:
        story_mode = kwargs['story_mode']
    
    # get a station list
    station_list = list(dataManager['weather'].read().keys())

    # get a translator
    t = config.translator(en=_TRANSLATE_EN, de=_TRANSLATE_DE)

    # switch the story_mode
    if not story_mode:
        st.session_state.selected_station = station_list[0]
        st.session_state.temporal_agg = "Annual"
        st.session_state.include_climate = False
    
    # we are in story mode! hurray!
    selected_station_selector(
        dataManager=dataManager,
        config=config,
        expander_container=expander_container,
        station_list=station_list,
        translator=t
    )
    temporal_agg_selector(
        dataManager=dataManager,
        config=config,
        expander_container=expander_container,
        translator=t
    )
    rcp_selector(
        dataManager=dataManager,
        config=config,
        expander_container=expander_container,
        translator=t
    )


def debug_main(**kwargs):
    from ruins.core import build_config
    params = st.experimental_get_query_params()
    config, dm = build_config(url_params=params, **kwargs)

    data_select(dataManager=dm, config=config)

    st.json(st.session_state)

if __name__ == '__main__':
    import fire
    fire.Fire(debug_main)



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns


def yrplot_hm(sr, ref=[1980, 2000], ag='sum', qa=0.95, cbar_title='Temperature anomaly (K)', cmx='coolwarm', cmxeq=True, li=False):
    """
    @deprecated - use stripes_heatmap.yrplot_hm instead
    
    """
    # plot of heatmap with monthyl and annual stripes
    yrs = sr.index.year.unique()
    dummy = np.zeros((len(yrs), 14)) * np.nan
    dummy = pd.DataFrame(dummy)
    dummy.index = yrs
    dummy.columns = ['J', 'F', 'M', 'A', 'M', 'J', 'J', 'A', 'S', 'O', 'N', 'D', '', 'Year']

    for i in sr.index:
        dummy.iloc[i.year - yrs[0], i.month - 1] = sr.loc[i]
    for i in yrs:
        if ag == 'sum':
            dummy.iloc[i - yrs[0], 13] = sr.loc[sr.index.year == i].sum()
            cmx = 'coolwarm_r'
        elif ag == 'min':
            dummy.iloc[i - yrs[0], 13] = sr.loc[sr.index.year == i].min()
        elif ag == 'max':
            dummy.iloc[i - yrs[0], 13] = sr.loc[sr.index.year == i].max()
        else:  # ag == 'mean'
            dummy.iloc[i - yrs[0], 13] = sr.loc[sr.index.year == i].mean()

    if ref == None:
        pass
    else:
        refx = dummy.loc[ref[0]:ref[1]].mean(axis=0)
        dummy = dummy - refx

    if cmxeq:
        vxU = dummy.abs().quantile(qa).quantile(qa)
        vxL = -1. * vxU
    else:
        vxU = dummy.quantile(qa).quantile(qa)
        vxL = dummy.quantile(1. - qa).quantile(1. - qa)

    if ag == 'sum':
        dummy.iloc[:, 13] = dummy.iloc[:, 13] / 12

    fig = plt.figure(figsize=(8,len(dummy)/15.))
    ax = sns.heatmap(dummy, cmap=cmx, vmin=vxL, vmax=vxU, cbar_kws={'label': cbar_title})

    if ref == None:
        pass
    else:
        ax.add_patch(
            plt.Rectangle((0, ref[0] - yrs[0]), 12, ref[1] - ref[0], fill=True, edgecolor='red', facecolor='gray', lw=2,
                          alpha=0.3, clip_on=False))
        ax.add_patch(
            plt.Rectangle((13, ref[0] - yrs[0]), 1, ref[1] - ref[0], fill=True, edgecolor='red', facecolor='gray', lw=2,
                          alpha=0.3, clip_on=False))
        ax.annotate('Reference period', (0.5, ref[1] - yrs[0] - 2), color='white', weight='bold', ha='left',
                    va='bottom', alpha=0.6)

    if type(li) == int:
        ax.axhline(li - yrs[0], color='k', ls='--', lw=1, alpha=0.5)
        ax.annotate(' >> observed', (12.5, li - yrs[0] - 0.5), color='k', ha='center', va='bottom', alpha=0.6,
                    rotation=90.)
        ax.annotate('modelled << ', (12.5, li - yrs[0] + 0.5), color='k', ha='center', va='top', alpha=0.6,
                    rotation=90.)
        # ax.add_patch(plt.Rectangle((0, li-yrs[0]), 12, 0, fill=False, edgecolor='k', ls='--', lw=1, alpha=0.5, clip_on=False))

    ax.set_ylabel('Year')
    ax.set_xlabel('Month          ')
    
    return fig


def monthlyx(dy, dyx=1, ylab='T (°C)', clab1='Monthly Mean in Year', clab2='Monthly Max in Year', pls='cividis_r') -> plt.Figure:
    cmap = plt.cm.get_cmap(pls)
    colors = cmap(np.linspace(0, 1, len(dy.index.year.unique())+1))
    colorsx = cmap(np.arange(cmap.N))

    idx1 = dy.index.year - dy.index.year.min()
    idx1m = dy.index.month

    if type(dyx) == int:
        pass
    else:
        idx2 = (dyx.index.year - dyx.index.year.min()).astype(int)
        idx2m = dyx.index.month

        cmap1 = plt.cm.get_cmap('gist_heat_r')
        colors1 = plt.cm.gist_heat_r(np.linspace(0, 1, len(dyx.index.year.unique()) + 1))
        colorsx1 = cmap1(np.arange(cmap1.N))

    for i in dy.columns:
        plt.scatter(idx1m + (np.random.rand(len(dy)) - 1.5), dy[i].values.astype(np.float), c=colors[idx1], alpha=0.6, s=2)

    if type(dyx) == int:
        pass
    else:
        for i in dyx.columns:
            plt.scatter(idx2m + (np.random.rand(len(dyx)) - 1.5), dyx[i].values.astype(np.float), c=colors1[idx2], alpha=0.6, s=2)

    cbar = plt.colorbar(plt.cm.ScalarMappable(cmap=cmap), label=clab1, ticks=[0, 0.2, 0.4, 0.6, 0.8, 1])
    cbar.ax.set_yticklabels(np.round(np.linspace(dy.index.year.min(), dy.index.year.max(), 6)).astype(int).astype(str))

    if type(dyx) == int:
        pass
    else:
        cbar1 = plt.colorbar(plt.cm.ScalarMappable(cmap=cmap1), label=clab2, ticks=[0, 0.2, 0.4, 0.6, 0.8, 1])
        cbar1.ax.set_yticklabels(
            np.round(np.linspace(dyx.index.year.min(), dyx.index.year.max(), 6)).astype(int).astype(str))

    plt.xticks([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
           ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'])
    plt.ylabel(ylab)

    fig = plt.gcf()
    return fig




import plotly.graph_objects as go
from scipy import stats
import numpy as np


BORDERS_COLORS = { # contains LOWER edges
    4: '#00A4B4',  # peacock blue
    7: '#00B5E2',  # sky blue
    9: '#C8E9E9',  # ice blue
    13: '#ffd7cb', # light salmon
    17: '#C63A4F', # watermelon
    20: '#800080'  # purple
}

BORDER_NAMES = ['extreme frost', 'frost', 'cold', 'warm', 'hot', 'tropic']


def plot_extreme_pdf(mus: float, stds: float, fig: go.Figure = None, border_colors=BORDERS_COLORS, border_names=BORDER_NAMES, x_range=(0, 25), x_res: int = 100, y_res: int = 25, label_margin_scale=1.1) -> go.Figure:
    """
    Temperature PDF shift plot.
    This function returns a plotly figure containing a heavily styled plotly PDF plot.
    It is controlled by passing the PDF moment(s) to the plot. If called with only one
    location and scale, the styled PDF is returned.
    If two locations and scales are passed, but both scales are the same, the PDF is
    only shifted by location and the original PDF is sketched. In case the scales differ,
    the new PDF will be correctly annotated.
    """
    # check input data
    if isinstance(mus, float):
        mu = mus
    else:
        mu = mus[1]

    if isinstance(stds, float):
        std = stds
    else:
        std = stds[1]

    # base data
    x = np.arange(*x_range, step=(x_range[1] - x_range[0]) / x_res)

    # get the distributions
    y = stats.norm.pdf(x, mu, std)
    y_max = y.max()
    if not isinstance(mus, float):
         # get the old PDF as well
        y_orig = stats.norm.pdf(x, mus[0], stds[0])
        
        # get the new Y max
        y_axis_max = max(y_orig.max(), y_max)
    else:
        y_axis_max = y_max
    
    # build the container for circles and circle colors
    c_x = []
    c_y = []
    colors = []

    # main, ugly loop
    for _x, _y in zip(x, y):
        # do not plot circles if they overlap with one of the mean markers
        if abs(_x - mu) < 1e-5:
            continue 
        if not isinstance(mus, float) and abs(_x - mus[0]) < 1e-5:
            continue

        # ratio of current prob.
        if _y / y_max < 0.01:
            continue
    
        # add circles - with y_res circles at max
        for t in np.arange(0, y_axis_max, y_axis_max / y_res):
            if t < _y * 0.95:
                c_x.append(_x)
                c_y.append(t)

                # check color
                # this is super ugly...
                c = border_colors[min(border_colors.keys())]
                for b, col in border_colors.items():
                    if _x > b:
                        c = col
                colors.append(c)
    
    # ugly part finished, do the plot
    if fig is None:
        fig = go.Figure()

    # handle if the shift already happend
    if not isinstance(mus, float):
        # plot the PDF and annotation
        fig.add_trace(
            go.Scatter(x=x, y=y_orig, mode='lines', line=dict(color='rgba(0,0,0,0.5)', dash='dash', width=2))
        )
        fig.add_trace(
            go.Scatter(x=[mus[0], mus[0]], y=[0, label_margin_scale * y_axis_max * 0.95], mode='lines', line=dict(color='rgba(0,0,0,0.5)', dash='dash'))
        )
        arr = dict(showarrow=True, arrowhead=3, arrowwidth=1.5, xref='x', yref='y', axref='x', ayref='y', standoff=8, startstandoff=6)
        fig.add_annotation(x=mus[0], y=label_margin_scale * y_axis_max, text='T<sub>0</sub>', showarrow=False, font=dict(size=20))
        fig.add_annotation(x=mus[1], y=label_margin_scale * y_axis_max, ax=mus[0], ay=label_margin_scale * y_axis_max, **arr)

        # check if the PDF changed the size
        if abs(stds[0] - stds[1] ) > 1e-5:
            fig.add_annotation(x=mu, y=y_max, ax=mu, ay=label_margin_scale * y_axis_max, **arr)
            pass
    else:
        # use current y_max as there is no second PDF
        y_axis_max = y_max

    # plot the color borders
    #for border, color in color_borders.items():
    #    fig.add_trace(
    #        go.Scatter(x=[border, border], y=[0, y_max], mode='lines', line=dict(dash='dash', color=color))
    #    )

    # main PDF
    fig.add_trace(
        go.Scatter(x=x, y=y, mode='lines', line=dict(color='black', width=3))
    )

    # colored circles
    fig.add_trace(
        go.Scatter(x=c_x, y=c_y, mode='markers', marker=dict(size=5, color=colors))
    )

    # handle temperature markers
    if not isinstance(stds, float) and stds[0] != stds[1]:
        mark_y = y_max
    else:
        mark_y = label_margin_scale * y_axis_max * 0.95
    fig.add_trace(
        go.Scatter(x=[mu, mu], y=[0, mark_y], mode='lines', line=dict(color='black', dash='dash'))
    )
    txt = f'T<sub>{0 if isinstance(mus, float) else 1}</sub>'
    fig.add_annotation(x=mu, y=label_margin_scale * y_axis_max, text=txt, showarrow=False, font=dict(size=20))
    
    # y axis annotations
    dom = dict(xref='x domain', yref='y domain', axref='x domain', ayref='y domain')
    fig.add_annotation(x=-0.01, y=0.95,ax=-0.01, ay=0.1, showarrow=True, arrowhead=2, arrowwidth=3, arrowcolor='rgba(0,0,0,0.5)', **dom)
    fig.add_annotation(x= -0.01, y=0.5, text='frequency', showarrow=False, textangle=270, font=dict(size=18, color='rgba(0,0,0,0.5)'), **dom)

    # x axis annotations
    for i, (low, up)  in enumerate(zip(list(border_colors.keys()), list(border_colors.keys())[1:] + [x_range[1]])):
        ref = dict(xref='x', yref='y', axref='x', ayref='y', startstandoff=1, standoff=1)
        ax = low if low != min(border_colors.keys()) else 0
        fig.add_annotation(x=up, y=-0.02, ax=ax, ay=-0.02, showarrow=True, arrowside='start+end', arrowhead=2, startarrowhead=2, arrowwidth=2, arrowcolor=border_colors[low], **ref)
        fig.add_annotation(x=ax + (up - ax) / 2, y=-0.035, showarrow=False, text=border_names[i], font=dict(size=16, color=border_colors[low]), xref='x', yref='y')

    # some layout fixes
    fig.update_layout(
        template='plotly_white',
        xaxis=dict(showgrid=False, zeroline=False, visible=False),
        yaxis=dict(showgrid=False, zeroline=False, visible=False),
        showlegend=False
    )
    
    # return
    return fig



from .maps import plt_map
from .kde import kde
from .weather_data import monthlyx
from .stripes_heatmap import yrplot_hm
from .climate_parcoords import climate_projection_parcoords
from .dists import distribution_plot
from .cordex_sunburst import sunburst
from .distribution_shift import plot_extreme_pdf


"""
Model scale selector
"""
import streamlit as st

from ruins.core import DataManager, Config


_TRANSLATE_EN = dict(
    title='Climate models have scale',
    description="""
There are many scientists working in different groups to build and enhance computer models of the climate physics.
These models operate at different scales in space and time. In our case, we refer to i) **global climate model** 
projections with a spatial resolution of approx. 120 km of one pixel at the surface and to ii) 
**regional climate model** projections with a spatial resolution of approx. 11 km.
As the name suggests, global climate models (or GCMs) cover the whole globe.
Their speciality is to simulate the overall physics of the atmosphere and land surface interactions.
The regional climate models (or RCMs) build on top of GCMs and build physical and/or statistical relationships
between GCM-outputs and locally observed weather. They seek to compensate the very coarse approximations of GCMs at
the price of more data being extrapolated into the future.

For more detailed reading, please continue here:
[CarbonBrief - How do climate models work?](https://www.carbonbrief.org/qa-how-do-climate-models-work)
    """,
    options={
        'Global': 'Global model',
        'Regional': 'Regional model'
    }
)

_TRANSATE_DE = dict(
    title='Skalierung von Klimamodellen',
    description="""
Es gibt unzählige Wissenschaftler_innen und Institutionen, die an einer Verbesserung der Computermodelle und dem
zugrundeliegenden physikalischen Verständnis arbeiten. Diese Modelle operieren auf verschiedenen *räumlichen* und 
*zeitlichen* Skalen. In dieser App haben **Globale Klimamodelle** eine räumlcihe Auflösung von ca. 120km pro Pixel an der
Erdoberfläche. **Regionale Klimamodelle** haben hingegen eine räumliche Auflösung von ca. 11km.
Dafür bedecken globale Modelle, wie der Name schon sagt, die ganze Welt. Sie sind besonders gut geeignet um globale Strömungen
und Interaktionen zwischen der Atmophäre und der Landoberfläche zu simulieren.
Die Regionalen modelle (RCMs) bauen dann weiter auf den (GCMs) auf, indem Sie globale Modelle entweder über statistische
oder physikalische Zusammenhänge mit lokal erhobenen Messdaten verknüpfen. Damit soll die grobe Auflösung von GCMs
wettgemacht werden.

Für mehr Informationen, lese hier weiter:
[CarbonBrief - Wie funktionieren Klimamodelle?](https://www.carbonbrief.org/qa-how-do-climate-models-work)
    """,
    options={
        'Global': 'Globales Modell',
        'Regional': 'Regionales Modell'
    }
)


def model_scale_selector(dataManager: DataManager, config: Config, expander_container=st.sidebar, **kwargs):
    """
    """
    # get the container
    container = st if 'container' not in kwargs else kwargs['container']

    # get a translator
    t = config.translator(en=_TRANSLATE_EN, de=_TRANSATE_DE)

    # get the translated options
    OPTIONS = t('options')
    # check if main page was already shown
    if config.has_key('climate_scale'):
        expander_container.radio(
            'Climate model' if config.lang == 'en' else 'Klimamodell',
            options=list(OPTIONS.keys()),
            format_func=lambda k: OPTIONS.get(k),
            key='climate_scale'
        )
        return

    # add the title
    container.title(t('title'))

    # add the description
    container.markdown(t('description'), unsafe_allow_html=True)

    # add the columns
    right, left = container.columns(2)

    right.info('THIS IS A THUMBNAIL')
    use_global = right.button('USE GLOBAL MODELS' if config.lang == 'en' else 'GLOBALES MODELL')

    left.info('THIS IS A THUMBNAIL')
    use_regional = left.button('USE REGIONAL MODELS' if config.lang == 'en' else 'REGIONALES MODELL')

    if use_global:
        st.session_state.climate_scale = 'Global'
        st.experimental_rerun()
    elif use_regional:
        st.session_state.climate_scale = 'Regional'
        st.experimental_rerun()
    else:
        st.stop()


def model_scale_select(dataManager: DataManager, config: Config, expander_container=st.sidebar, **kwargs):
    """
    """
    # set default story mode 
    story_mode = config.get('story_mode', True)
    if 'story_mode' in kwargs:
        story_mode = kwargs['story_mode']

    # pre-select 
    if config.has_key('climate_scale') or not story_mode:
        st.session_state.climate_scale = config.get('climate_scale', 'Global')
    
    # call the seletor
    model_scale_selector(dataManager, config, expander_container=expander_container)


def debug_main(**kwargs):
    from ruins.core import build_config
    params = st.experimental_get_query_params()
    config, dataManager = build_config(url_params=params, **kwargs)

    model_scale_select(dataManager, config)

    st.json(st.session_state)


if __name__ == '__main__':
    import fire
    fire.Fire(debug_main)



import streamlit as st
import pandas as pd

from ruins.core import build_config, debug_view, DataManager, Config
from ruins.plotting import bubble_plot


_TRANSLATE_EN = dict(
    title='Bubble plot',
    introduction="""
    Bubble plot of RCP scenarios.
"""
)

_TRANSLATE_DE = dict(
    title='Bubble plot',
    introduction="""
    Bubble plot der RCP Szenarien.
"""
)


def concept_explainer(config: Config, **kwargs):
    """Show an explanation, if it was not already shown.
    """
    # check if we saw the explainer already
    if config.has_key('bubbles_explainer'):
        return
    
    # get the container and a translation function
    container = kwargs['container'] if 'container' in kwargs else st
    t = config.translator(en=_TRANSLATE_EN, de=_TRANSLATE_DE)

    # place title and intro
    container.title(t('title'))
    container.markdown(t('introduction'), unsafe_allow_html=True)

    # check if the user wants to continue
    accept = container.button('WEITER' if config.lang == 'de' else 'CONTINUE')
    if accept:
        st.session_state.bubbles_explainer = True
        st.experimental_rerun()
    else:
        st.stop()


def show_bubbles(dataManager: DataManager, config: Config):
    """
    Shows the bubble plot.
    """
    data = dataManager['cordex_krummh_nobias_chk_f32_ET'].read()

    possibles = list(data.columns)
    possibles[0] = 'None'

    level1 = st.sidebar.radio('Select Index1:', possibles, index = 1)
    level2 = st.sidebar.radio('Select Index2:', possibles, index = 2)
    level3 = st.sidebar.radio('Select Index3:', possibles, index = 0)

    if(level1 == possibles[0]):
        select = []
    else:
        if (level2 == possibles[0]):
            select = [level1]
        else:
            if (level3 == possibles[0]):
                select = [level1, level2]
            else:
                select = [level1, level2, level3]

    fig = bubble_plot.draw_bubbles(data, selectors = select)

    st.pyplot(fig)
    

def main_app(**kwargs):
    """
    """
    # build the config and dataManager from kwargs
    url_params = st.experimental_get_query_params()
    config, dataManager = build_config(url_params=url_params, **kwargs)

    # set page properties and debug view    
    st.set_page_config(page_title='Bubble plot', layout=config.layout)
    debug_view.debug_view(dataManager, config, debug_name='DEBUG - initial state')

    # explainer
    concept_explainer(config)

    # show bubbles
    show_bubbles(dataManager, config)

    # end state debug
    debug_view.debug_view(dataManager, config, debug_name='DEBUG - finished app')


if __name__ == '__main__':
    import fire
    fire.Fire(main_app)



import numpy as np
import circlify
import matplotlib.pyplot as plt


def draw_bubbles(a, selectors = ['RCP','GCM']):
    ##############################
    # Erstellen der Struktur für das circlelize package - geht bestimmt viel einfacher und eleganter und für mehr Ebenen.

    # Die Struktur wird z.B. auch in einem pandas multi-Index gut wieder gegeben, aber wie überführt man die in die hierarische Liste???
    #tab = a.groupby(['RCP','GCM','RCM','ensemble','institute']).size()
    #tab

    levels = len(selectors)
    
    data = [{'id': 'climate', 'datum': a.shape[0]}]

    z = 0
    b = a[selectors[z]].unique()
    listed = []
    for bb in b : listed = np.append(listed,{'id' : bb, 'datum' : a.groupby([selectors[z]]).size()[bb]})
    data[0]['children'] = list(listed)

    if levels >1:

        for bb in data[0]['children']:
            z = 1
            ax = a[a[selectors[z-1]]==bb['id']]
            bx = ax[selectors[z]].unique()
            listed = []
            for bbx in bx : listed = np.append(listed,{'id' : bbx, 'datum' : ax.groupby([selectors[z]]).size()[bbx]})
            bb['children'] = list(listed)

            if levels >2:
                for cc in bb['children']:
                    cx = ax[ax[selectors[z]]==cc['id']]
                    ccx = cx[selectors[z+1]].unique()
                    listedc = []
                    for cccx in ccx : listedc = np.append(listedc,{'id' : cccx, 'datum' : ax.groupby([selectors[z+1]]).size()[cccx]})
                    cc['children'] = list(listedc)


    ####################
    # Berechnen der Kreisgeometrien

    # Compute circle positions thanks to the circlify() function
    circles = circlify.circlify(
        data, 
        show_enclosure=False, 
        target_enclosure=circlify.Circle(x=0, y=0, r=1)
    )
        
    #################### 
    # eigentliches Erstellen des Circle Plots

    # Create just a figure and only one subplot
    fig, ax = plt.subplots(figsize=(14,14))

    # Title
    ax.set_title('Repartition of the climate predictions')

    # Remove axes
    ax.axis('off')

    # Find axis boundaries
    lim = max(
        max(
            abs(circle.x) + circle.r,
            abs(circle.y) + circle.r,
        )
        for circle in circles
    )
    plt.xlim(-lim, lim)
    plt.ylim(-lim, lim)

    # Print circle the highest level:
    for circle in circles:
        if circle.level != 2:
            continue
        x, y, r = circle
        ax.add_patch( plt.Circle((x, y), r, alpha=0.5, linewidth=2, color="lightblue"))

    # Print circle and labels for the highest level:
    for circle in circles:
        if circle.level != 3:
            continue
        x, y, r = circle
        label = circle.ex["id"]
        ax.add_patch( plt.Circle((x, y), r, alpha=0.5, linewidth=2, color="darkblue"))
        plt.annotate(label, (x,y-r/2), ha='center', color="white")

    # Print circle and labels for the lowest level:
    for circle in circles:
        if circle.level != 4:
            continue
        x, y, r = circle
        label = circle.ex["id"]
        ax.add_patch( plt.Circle((x, y), r, alpha=0.5, linewidth=2, color="#69b3a2"))
        plt.annotate(label, (x,y ), ha='center', color="white")

    # Print labels for the highest level:
    for circle in circles:
        if circle.level != 2:
            continue
        x, y, r = circle
        label = circle.ex["id"]
        plt.annotate(label, (x,y-r ) ,va='center', ha='center', bbox=dict(facecolor='white', edgecolor='black', boxstyle='round', pad=.5))
        
    plt.annotate('No. of Elements:'+str(a.shape[0]), (-0.9,0.9 ))

    return fig


import streamlit as st
import numpy as np

import pandas as pd # TODO  this should be covered by the DataManager
import xarray as xr # TODO  this should be covered by the DataManager
import matplotlib.pyplot as plt # TODO this should be moved into plotting submodule


from ruins import components


def load_alldata():
    weather = xr.load_dataset('data/weather.nc')
    climate = xr.load_dataset('data/cordex_coast.nc')

    # WARNING - bug fix for now:
    # 'HadGEM2-ES' model runs are problematic and will be removed for now
    # The issue is with the timestamp and requires revision of the ESGF reading routines
    kys = [s for s in list(climate.keys()) if 'HadGEM2-ES' not in s] #remove all entries of HadGEM2-ES (6 entries)
    climate = climate[kys]

    return weather, climate


def applySDM(wdata, data, meth='rel', cdf_threshold=0.9999999, lower_limit=0.1):
    '''apply structured distribution mapping to climate data and return unbiased version of dataset'''
    from sdm import SDM
    data_ub = data.copy()

    for k in data_ub.columns:
        data_col = data_ub[k].dropna()
        overlapx = pd.concat(
            [wdata.loc[data_col.index[0]:wdata.index[-1]], data_col.loc[data_col.index[0]:wdata.index[-1]]], axis=1)
        overlapx.columns = ['obs', 'cm']
        overlapx = overlapx.dropna()
        try:
            data_ub[k] = SDM(overlapx.obs, overlapx.cm, data_col, meth, cdf_threshold, lower_limit)
        except:
            data_ub[k] = data_ub[k] * np.nan

    data_ub[data_ub == 0.0000000] = np.nan
    data_ub = data_ub.loc[data_ub.index[0]:pd.to_datetime('2099-12-31 23:59:59')]

    return data_ub



def ub_climate(cdata, wdata, ub=True):
    varis = ['T', 'Tmax', 'Tmin', 'aP', 'Prec', 'RH', 'Rs', 'u2', 'EToHG']
    firstitem = True
    for vari in varis:
        data = cdata.sel(vars=vari).to_dataframe()
        data = data[data.columns[data.columns != 'vars']]

        if (vari == 'T') | (vari == 'Tmax') | (vari == 'Tmin'):
            meth = 'abs'
        else:
            meth = 'rel'

        if ub:
            wdatax = wdata.sel(vars=vari).to_dataframe().iloc[:, -1].dropna()
            data_ub = applySDM(wdatax, data, meth=meth)

        else:
            data_ub = data

        data_ubx = data_ub.mean(axis=1)
        data_ubx.columns = [vari]

        if firstitem:
            data_ubc = data_ubx
            firstitem = False
        else:
            data_ubc = pd.concat([data_ubc, data_ubx], axis=1)

    data_ubc.columns = varis
    return data_ubc


def get_turbine(pfi, plotit=False):
    dummy = pd.read_csv(pfi)
    v_start = dummy.loc[3].astype(np.float).values[0]
    v_max = dummy.loc[2].astype(np.float).values[0]
    P_v = dummy.loc[4:28].astype(np.float)
    P_v.index = np.arange(int(np.ceil(v_max)) + 1)[1:]
    if plotit:
        P_v.plot()
    return [P_v, v_start, v_max]


def P_wind(wind, pfi):
    [P_v, v_start, v_max] = get_turbine(pfi, False)

    # stephours = (wind.index[2] - wind.index[1]).days * 24
    # stephours = 1.

    def interp(val):
        if (val >= v_max) | (val < v_start):
            return 0.
        elif ~np.isnan(val):
            if np.ceil(val) == np.floor(val):
                return P_v.loc[int(np.floor(val))].values[0]
            else:
                x1 = P_v.loc[int(np.floor(val))].values[0]
                x2 = P_v.loc[int(np.ceil(val))].values[0]
                return ((val - np.floor(val)) * x2 + (1. - (val - np.floor(val))) * x1)
        else:
            return np.nan

    ip_vec = np.vectorize(interp)

    def get_windpower(u2):
        val = np.fmax(-1. * np.cos(np.arange(0., 2 * np.pi, 0.27)) + u2, 0.)
        return np.sum(ip_vec(val))

    return wind.apply(get_windpower)


def cropmodel(data,crop='wheat',rcp='rcp85',name='croprunname'):
    import get_climate_data.cli_crop as clc
    #read co2 concentrations
    if rcp == 'rcp85':
        CO2 = pd.read_csv('data/RCP85_MIDYEAR_CONCENTRATIONS.DAT', skiprows=38, delim_whitespace=True, index_col=0).CO2EQ
    elif rcp == 'rcp45':
        CO2 = pd.read_csv('data/RCP45_MIDYEAR_CONCENTRATIONS.DAT', skiprows=38, delim_whitespace=True, index_col=0).CO2EQ
    else:
        CO2 = pd.read_csv('data/RCP45_MIDYEAR_CONCENTRATIONS.DAT', skiprows=38, delim_whitespace=True, index_col=0).CO2EQ * 0. + 400.

    if crop == 'maize':
        x = np.array([9.375e+00, 3.198e+01, 1.973e+00, 8.700e+01, 1.144e+01, 3.630e+01, 7.260e-02, 1.237e+00, 2.180e+03, 1.501e-01, 5.230e-01, 5.678e+01, 6.970e+02])
    elif crop == 'meadow':
        x = np.array([6.543e+00, 1.238e+01, 1.029e+00, 8.706e+01, 1.510e+01, 3.253e+01, 1.199e+00, 1.535e+03, 7.784e+03, 6.530e+03, 8.030e+03, 8.092e+03, 5.884e+03])
    elif crop == 'wheat':
        x = np.array([1.257e+00, 1.276e+01, 1.101e+00, 1.010e+02, 2.578e+01, 2.769e+01, 3.416e-01, 4.940e-01, 1.906e+03, 1.921e-01, 4.595e-01, 6.066e+01, 5.360e+02])
    else:
        print('ERROR: Crop not specified with parameters.')

    yields = clc.cli_SC(data, x, CO2, nme=name)

    return yields


def water_proj():
    #hsim = pd.read_csv('data/hsim.csv',index_col=0)
    #hsim.index = pd.to_datetime(hsim.index)
    hsim_collect = pd.read_csv('data/hsim_collect.csv', index_col=0)
    hsim_collect.index = pd.to_datetime(hsim_collect.index)
    all_vals = np.ravel(hsim_collect.values)[~np.isnan(np.ravel(hsim_collect.values))]

    perci = st.slider('Adjust percentile of extreme events:', min_value=90., max_value=99.9999, value=99., key='perci')

    firstitem = True
    for i in hsim_collect.columns:
        hc = hsim_collect.loc[hsim_collect[i] > np.percentile(all_vals, perci), i].resample('1Y').count()
        hc.name = i
        if firstitem:
            h_count = hc
            firstitem = False
        else:
            h_count = pd.concat([h_count, hc], axis=1)

    fig = plt.figure(constrained_layout=True,figsize=(10,2.5))
    gs = fig.add_gridspec(1, 5)
    ax = fig.add_subplot(gs[0, :-1])
    ax1 = ax.twinx()
    ax2 = fig.add_subplot(gs[0, -1])

    all_vals = np.ravel(hsim_collect.values)[~np.isnan(np.ravel(hsim_collect.values))]
    sns.distplot(all_vals, ax=ax2)
    ax2.vlines(np.percentile(all_vals, perci), 0., 0.06, colors='red')
    ax2.text(np.percentile(all_vals, perci), 0.065, str(perci) + '%', c='red', ha='center')
    ax2.set_xlim(-3, np.percentile(all_vals, 99.98))

    for i in hsim_collect.columns:
        hsim_collect[i].plot(style='.', label='_nolegend_', c='gray', alpha=0.2, ax=ax)
        try:
            hsim_collect.loc[hsim_collect[i] > np.percentile(all_vals, perci), i].plot(style='.', label='_nolegend_',
                                                                                   c='red', alpha=0.2, ax=ax)
        except:
            pass

    (h_count.sum(axis=1) / len(h_count.columns)).rolling(7).mean().plot(ax=ax1, grid=False, ls='--', label='all')
    cc = [x for x in h_count.columns if x.split('.')[-1] == 'rcp45']
    (h_count[cc].sum(axis=1) / len(cc)).rolling(7).mean().plot(ax=ax1, grid=False, label='rcp45')

    cc = [x for x in h_count.columns if x.split('.')[-1] == 'rcp85']
    (h_count[cc].sum(axis=1) / len(cc)).rolling(7).mean().plot(ax=ax1, grid=False, label='rcp85')
    ax1.legend()
    ax.set_ylabel('Q (mm/day)')
    ax1.set_ylabel('days w/ Q>' + str(np.round(perci / 100., 3)) + ' percentile\n7 year rolling mean')
    st.pyplot()



def management_explorer(w_topic: str):
    rcps = ['rcp45', 'rcp85']  # 'rcp26',
    rcp = st.sidebar.selectbox('RCP:', rcps)

    weather, climate = load_alldata()
    data = climate.filter_by_attrs(RCP=rcp)

    ub = st.sidebar.checkbox('Apply SDM bias correction', True)

    cdata = ub_climate(data, weather['krummhoern'], ub)

    if w_topic == 'Agriculture':
        crops = ['wheat', 'maize', 'meadow']
        crop = st.sidebar.selectbox('Crop:', crops)

        yieldx = cropmodel(cdata,crop,rcp,name=crop+' @ '+rcp)

        yieldx.rolling(7).mean().plot()
        st.pyplot()

    elif w_topic == 'Wind Energy':
        storeddata = st.checkbox('Use pre-calculated data (calculation not yet optimized - very slow)', True)
        if storeddata:
            cwind_p = pd.read_csv('data/windpowerx.csv', index_col=0)
            cwind_p.index = pd.to_datetime(cwind_p.index)
            plt.figure(figsize=(10, 2.5))
            cwind_p[[x for x in cwind_p.columns if rcp in x][::3]].mean(axis=1).plot(label='Enercon 3MW')
            cwind_p[[x for x in cwind_p.columns if rcp in x][1::3]].mean(axis=1).plot(label='9x Enercon 0.33MW')
            cwind_p[[x for x in cwind_p.columns if rcp in x][2::3]].mean(axis=1).plot(label='0.4x Enercon 7.5MW')
        else:
            wpengine = ['Enercon 3MW', 'Vestas 3MW', '9x Enercon 0.33MW', '13x Vestas 0.23MW', '0.4x Enercon 7.5MW']
            #wpengine = ['Enercon 3MW', '9x Enercon 0.33MW', '0.4x Enercon 7.5MW']
            turbine = st.sidebar.selectbox('Select turbine:', wpengine)

            plt.figure(figsize=(10,2.5))
            if turbine == 'Enercon 3MW':
                pfi = 'data/pow/Enercon E-115 3000kW (MG).pow'
                P_wind(cdata.u2, pfi).resample('1Y').sum().rolling(7).mean().plot(label=turbine)
            elif turbine == 'Vestas 3MW':
                pfi = 'data/pow/Vestas 112m 3MW (MT).pow'
                P_wind(cdata.u2, pfi).resample('1Y').sum().rolling(7).mean().plot(label=turbine)
            elif turbine == '9x Enercon 0.33MW':
                pfi = 'data/pow/Enercon E33_33.4m _330kw(MT).pow'
                (P_wind(cdata.u2, pfi).resample('1Y').sum()*9.).rolling(7).mean().plot(label=turbine)
            elif turbine == '13x Vestas 0.23MW':
                pfi = 'data/pow/Vestas V29-225kw(MT).pow'
                (P_wind(cdata.u2, pfi).resample('1Y').sum() * 13.).rolling(7).mean().plot(label=turbine)
            elif turbine == '0.4x Vestas 7.5MW':
                pfi = 'data/pow/Enercon E-126_127m_7500kW (MT).pow'
                (P_wind(cdata.u2, pfi).resample('1Y').sum() * 0.4).rolling(7).mean().plot(label=turbine)

        plt.legend()
        plt.ylabel('Wh/yr')
        st.pyplot()

    elif w_topic == 'Extreme Events':
        water_proj()


def main_app():
    st.header('Management Explorer')
    st.markdown('''In this section we project the climate model outputs to ecosystem services through diffferent model approaches. The impact of potential management is tackeled at a rather broad level in terms of exemplary options.''', unsafe_allow_html=True)
    
    # TODO: refactor this
    topics = ['Warming', 'Weather Indices', 'Drought/Flood', 'Agriculture', 'Extreme Events', 'Wind Energy']
    
    # topic selector
    topic = components.topic_selector(topic_list=topics, container=st.sidebar)
    
    # TODO refactor this    
    management_explorer(topic)


if __name__ == '__main__':
    main_app()



import streamlit as st

from ruins.core import Config, DataManager


def debug_view(dataManager: DataManager, config: Config, debug_name: str = None) -> None:
    '''
    Set Config['debug'] = 'True' to display debug view which
    shows current Config and dataManager parameters.
    '''
    if config.debug:
        name = f'DEBUG [{debug_name}]' if debug_name else 'DEBUG'
        exp = st.expander(name, expanded=True)
        left, right = exp.columns(2)

        left.markdown('## Config paramters')
        left.json(dict(config))

        right.markdown('## Session state')
        right.json(st.session_state)



from typing import List

import streamlit as st
import pandas as pd

from ruins.core import build_config, debug_view, Config, DataManager
from ruins.plotting import sunburst
from ruins.processing.sunburst import ordered_sunburst_data


_TRANSLATE_DE = dict(
    title='Verwendete Klimamodelle',
    intro="""Es gibt eine Vielzahl an unterschiedlichen Klimamodellen,
die auf unterschiedlichen Skalen arbeiten und von unterschiedlichen Voraussetzungen ausgehen.
Das macht es schwer von **dem Klimamodell** zu sprechen und reflektiert die 
Unsicherheit über die Zukunft.

In der Abbildung unten findest du alle Modelle, die für die Erstellung unserer 
Ergebnisse herangezogen wurden, kategorisiert nach dem Globalen und regionalen Modell, sowie
dem RCP Szenario.
"""
)

_TRANSLATE_EN = dict(
    title='Climate models in use',
    intro="""There are tons of different climate models, which operate at different
scales and have different underlying assumptions. That reflects the uncertainty about
our future and thus, there is not **the one correct climate model**.

All models used to calculate our results are shown in the chart. We categorized them by
the global and regional model in use as well as the RCP scenario.
"""
)


def plot_controls(config: Config, expander=st.sidebar) -> None:
    """Add the controls to the application"""
    with expander.expander('Hierachie' if config.lang=='de' else 'Hierachy', expanded=True):
        # set the order 
        o = st.radio('Reihenfolge' if config.lang=='de' else 'Order', options=['GCM -> RCM -> RCP', 'RCP -> GCM -> RCM', 'RCM -> GCM -> RCP'])
        st.session_state.sunburst_order = o.split(' -> ')

        # set the level
        LVL = {2: '1 level', 3: '2 levels', 4: '3 levels'}
        st.select_slider(
            'Anzahl Stufen' if config.lang=='de' else 'Show levels',
            options=list(LVL.keys()),
            value=4,
            format_func=lambda k: LVL.get(k),
            key='sunburst_maxdepth'
        ) 


@st.experimental_memo
def get_cached_data(_dataManager: DataManager, order: List[str]) -> pd.DataFrame:
    return ordered_sunburst_data(_dataManager, order)


def sunburst_plot(dataManager: DataManager, config: Config):
    """Add the plot"""
    # get the ordered data
    df = get_cached_data(dataManager, config['sunburst_order'])
    
    # get the figure
    fig = sunburst(df, maxdepth=config.get('sunburst_maxdepth', 4))

    # add
    st.plotly_chart(fig, use_container_width=True)


def main_app(**kwargs):
    """"""
    # build the config and dataManager from kwargs
    url_params = st.experimental_get_query_params()
    config, dataManager = build_config(url_params=url_params, **kwargs)

    # set page properties and debug view    
    st.set_page_config(page_title='Climate Model Sunburst', layout=config.layout)
    debug_view.debug_view(dataManager, config, debug_name='DEBUG - initial state')

    # get a translator
    t = config.translator(de=_TRANSLATE_DE, en=_TRANSLATE_EN)

    st.title(t('title'))
    st.markdown(t('intro'))

    # main application
    plot_controls(config)
    sunburst_plot(dataManager, config)

    # end state debug
    debug_view.debug_view(dataManager, config, debug_name='DEBUG - finished app')

if __name__=='__main__':
    import fire
    fire.Fire(main_app)



from typing import Tuple, List

import streamlit as st
import plotly.graph_objects as go
import plotly.express as px
from plotly.express.colors import unlabel_rgb, label_rgb, n_colors, sequential
import numpy as np
from scipy.stats import norm

from ruins.core import Config, build_config, debug_view
from ruins.plotting import distribution_plot

_TRANSLATE_EN = dict(
    title='Uncertainty & Risk',
    introduction="""
This first playground app will illustrate how uncertainty and risk influence our everyday decisions.
It is quite important to understand the difference between [knightian uncertaintry](https://en.wikipedia.org/wiki/Knightian_uncertainty)
and risk at this simplified example, before we move to climate modeling and weather data.
Taking the whole earth system into account, these concepts apply and are of high importance,
but their interpretration is way more complicated.

DESCRIPTION OF THE SIMPLIFIED EXAMPLE
    """,
    event_1_desc="""
The first event has two possible outcomes, from which we know their expected return distributions,
but we lack knowledge about the probabilities of occurence. Use the slider below to adjust their 
mean outcome and the deviations from this mean.
""",
    event_2_desc="""
In the first event, there was an uncalculatable uncertainty, as we can't predict the outcome. Instead of
throw the dice every time and risk being way off in terms of return, we can take a save route and make
an active decision for another event. The second event has only one outcome, from which we know that the
expected return is worse than the better outcome of the first event. 
We trade off the posibility of very positive outcome at the cost of not being trapped into very bad outcomes.

But is that worth it?
"""
)

_TRANSLATE_DE = dict(
    title='Unsicherheit und Risiko',
    introduction="""
Dieses erste, vereinfachte Beispiel demonstriert wie sich Unsicherheit und Risiko auf
unsere alltäglichen Entscheidungen auswirkt. Es ist wichtig den Unterschied zwischen 
[Knightsche Unsicherheit](https://de.wikipedia.org/wiki/Knightsche_Unsicherheit) anhand dieses stark vereinfachten
Beispiels zu erkunden und zu verstehen, bevor wir die Modelle und Daten der letzten Kapitel betrachten.
Betrachtet man das gesamte Erdsystem, sind Unsicherheit und Risiko für die Interpretierbarkeit der Daten
von fundamentaler Bedeutng, stellen sich jedoch in wesentlich komplexeren Zusammenhängen dar.

BESCHREIBUNG DES VEREINFACHTEN BEISPIELS
""",
    event_1_desc="""
Das erste Ereignis hat zwei verschiedene Ergebnisse. Für jedes kennen wir die drchschnittliche Erwartung und deren
Verteilung, allerdings haben wir keine Information über die Wahrscheinlichkeit, dass eines der Ergebnisse eintritt.
Benutze die Schieberegler um die Verteilungen der Ergebnisse anzupassen.
""",
    event_2_desc="""
Im ersten Ereignis mussten wir mit einer unbestimmbaren Unsicherheit umgehen, da wir das Ergebnis nicht vorhersagen konnten.
Anstatt hier die Würfel entscheiden zu lassen und ein schlechtes Ergebnis zu riskieren, können wir uns gänzlich
umentscheiden und Ereignis 2 eintreten lassen, das nur ein einziges Ergebnis hat. Allerdings ist das Ergebnis hier
schlechter als der bessere Ausgang des ersten Erignisses. Wir erkaufen uns die Sicherheit keine sehr schlechten Ergbnisse
zu haben damit, dass wir auch auf sehr positive Ergebnisse verzichten.

Aber lohnt sich das?
"""
)

def concept_explainer(config: Config, **kwargs):
    """Show an explanation, if it was not already shown.
    """
    # check if we saw the explainer already
    if config.has_key('uncertainty_playground_explainer'):
        return
    
    # get the container and a translation function
    container = kwargs['container'] if 'container' in kwargs else st
    t = config.translator(en=_TRANSLATE_EN, de=_TRANSLATE_DE)

    # place title and intro
    container.title(t('title'))
    container.markdown(t('introduction'), unsafe_allow_html=True)

    # check if the user wants to continue
    accept = container.button('WEITER' if config.lang == 'de' else 'CONTINUE')
    if accept:
        st.session_state.uncertainty_playground_explainer = True
        st.experimental_rerun()
    else:
        st.stop()


def _helper_plot(ev1: List[Tuple[float, float]], ev2: Tuple[float, float] = None, **kwargs) -> go.Figure:
    # create figure
    fig = go.Figure()

    # build the colorscale with enough colors
    cscale = getattr(sequential, kwargs.get('colorscale', 'Greens'))
    cmap = n_colors(unlabel_rgb(cscale[-1]), unlabel_rgb(cscale[-3]), len(ev1))

    # get a common x
    x = np.linspace(0, 10, 200)

    # iterate over all outcomes
    for i,outcome in enumerate(ev1):
        mu, std = outcome
        y = norm.pdf(x, loc=mu, scale=std)
        y_sum = y.sum()
        y /= y_sum

        # add the traces
        fig.add_trace(
            go.Scatter(x=x, y=y * 100, mode='lines', line=dict(color=label_rgb(cmap[i])), name=f'Outcome #{i + 1}', fill='tozerox')
        )
        fig.add_trace(
            go.Scatter(x=[mu, mu], y=[0, norm.pdf(mu, loc=mu, scale=std) / y_sum * 100], mode='lines', line=dict(color=label_rgb(cmap[i]), width=3, dash='dash'), name=f'Mean #{i + 1}')
        )
    
    # handle second event
    if ev2 is not None:
        mu, std = ev2
        y = norm.pdf(x, loc=mu, scale=std)
        y_sum = y.sum()
        y /= y_sum
        
        # add distribution
        fig.add_trace(
            go.Scatter(x=x, y=y * 100, mode='lines', line=dict(color='orange', width=2), name='Alternative Event', fill='tozerox')
        )
        # add mean
        fig.add_trace(
            go.Scatter(x=[mu, mu], y=[0, norm.pdf(mu, loc=mu, scale=std) / y_sum * 100], mode='lines', line=dict(color='orange', width=2, dash='dash'), name='Alternative Event mean value')
        )

    # adjust figure
    fig.update_layout(
        template='plotly_white',
        legend=dict(orientation='h')
    )

    return fig


def concept_graph(config: Config, expander_container=st.sidebar, **kwargs) -> go.Figure:
    """
    # TODO: document this
    """
    # get the container and translator
    container = kwargs['container'] if 'container' in kwargs else st
    t = config.translator(de=_TRANSLATE_DE, en=_TRANSLATE_EN)

    # ------------------------
    # First PDF
    if not config.has_key('concept_event_1'):
        container.markdown(t('event_1_desc'))
        l1, c1, r1 = container.columns(3)
        l2, c2, r2 = container.columns(3)

        # outcome 1
        l1.markdown('### Outcome 1')
        ou1_mu = c1.slider('Expected value of outcome #1', min_value=1., max_value=10., value=2.5)
        ou1_st = r1.slider('Certainty of outcome #1', min_value=0.1, max_value=3.0, value=0.5)
        
        # outcome 2
        l2.markdown('### Outcome 2')
        ou2_mu = c2.slider('Expected value of outcome #2', min_value=1., max_value=10., value=6.0)
        ou2_st = r2.slider('Certainty of outcome #2', min_value=0.1, max_value=3.0, value=0.4)

        ev1 = [(ou1_mu, ou1_st), (ou2_mu, ou2_st)]
        # add the continue button
        ev1_ok = container.button('WEITER' if config.lang=='de' else 'CONTINUE')
        if ev1_ok:
            st.session_state.concept_event_1 = ev1
            st.experimental_rerun()
        else:
            fig = distribution_plot({'outcomes': ev1, 'coloscale': 'Greens'})
            return fig
    else:
        ev1 = config['concept_event_1']
        ev1_new = []
        for i, out in enumerate(ev1):
            e = expander_container.expander(f'Outcome #{i + 1}', expanded=True)
            mu = e.slider(f'Expected value of outcome # {i + 1}', min_value=1., max_value=10., value=out[0])
            std = e.slider(f'Certainty of outcome # {i + 1}', min_value=0.1, max_value=2.0, value=out[1])
            ev1_new.append((mu, std, ))

    # ------------------------
    # add second event
    container.markdown(t('event_2_desc'))
    l, c, r = container.columns(3)

    # second event
    l.markdown('### Second event')
    e2_mu = c.slider('Expected value of alternative event', min_value=1., max_value=10., value=5.5)
    e2_st = r.slider('Certainty of alternative event', min_value=0.1, max_value=3.0, value=0.2)

    fig = distribution_plot({'outcomes': ev1_new, 'name': 'Original Event', 'colorscale': 'Greens'}, {'outcomes': [(e2_mu, e2_st)], 'name': 'Alternative Event', 'colorscale': 'Oranges'})
    return fig




def concept_playground(config: Config) -> None:
    """
    The concept playground demonstrates how knightian uncertainty
    is different from risk and how it influences everyday decisions.
    """
    # TODO: add the story mode stuff here

    # explainer
    concept_explainer(config)

    # show the graph
    
    fig = concept_graph(config)
    plot_area = st.empty()
    plot_area.plotly_chart(fig, use_container_width=True)


def main_app(**kwargs):
    """
    """
    # build the config and the dataManager from kwargs
    url_params = st.experimental_get_query_params()
    config, dataManager = build_config(url_params=url_params, **kwargs)

    # set page config and debug view
    st.set_page_config(page_title='Uncertainty Explorer', layout=config.layout)
    debug_view.debug_view(dataManager, config, debug_name='Initial Application State')

    # --------------------------
    # Main App

    # TODO: right now, we have only the playground here
    concept_playground(config)



if __name__ == '__main__':
    main_app()



from faulthandler import disable
from typing import List, Callable, Tuple
import streamlit as st
import xarray as xr     # TODO: these references should be moved to DataManager
import pandas as pd     # TODO: these references should be moved to DataManager
import numpy as np
import matplotlib.pyplot as plt
from plotly.express.colors import named_colorscales

from ruins.plotting import plt_map, kde, yrplot_hm, sunburst
from ruins.components import data_select, model_scale_select
from ruins.core import build_config, debug_view, DataManager, Config
from ruins.core.cache import partial_memoize
from ruins.plotting.climate_parcoords import climate_projection_parcoords
from ruins.processing.sdm import SDM


_TRANSLATE_DE_CLIMATE = dict(
    title="Unser Wetter in der Zukunft",
    introduction="""Schon die Betrachtung von aktuellen Wetterdaten zeigt, dass diese nicht immer eindeutig sind.
Unsicherheiten bei der Beobachtung von Wetterphänomenen Unwissen über Prozesse wirken sich
auf aktuellen Wetterbeschreibungen aus und erschweren z.B. die Wettervorhersage.

In der Klimaforschung müssen wir nun die nächsten **80 Jahre** vorhersagen. Neben Vorhersageunsicherheiten,
sind aber auch die **Voraussetzungen und Bedingungen** von denen wir ausgehen müssen nicht bekannt und 
auch nicht zu beziffern. 

Wie entwickelt sich die Weltbevölkerung? 

Wie viel CO2 stoßen wir in 30 Jahren aus?

Außerdem: Der Mensch wird durch Managemententscheidungen jene Bedingungen ständig ändern. Dennoch müssen wir heute 
schon Informationen bereitstellen und Abschätzungen liefern. Deshalb ist es so wichtig Wissenslücken zu
schließen und noch wichtiger, Unsicherheit durch Unwissen in Managaementprozesse einzubinden.

Mit dem **Klimamodell Explorer** können einige dieser Szenarien mit einander verglichen werden.
"""
)

_TRANSLATE_EN_CLIMATE = dict(
    title="Projecting weather into the future",
    introduction="""Schon die Betrachtung von aktuellen Wetterdaten zeigt, dass diese nicht immer eindeutig sind.
Unsicherheiten bei der Beobachtung von Wetterphänomenen Unwissen über Prozesse wirken sich
auf aktuellen Wetterbeschreibungen aus und erschweren z.B. die Wettervorhersage.

In der Klimaforschung müssen wir nun die nächsten **80 Jahre** vorhersagen. Neben Vorhersageunsicherheiten,
sind aber auch die **Voraussetzungen und Bedingungen** von denen wir ausgehen müssen nicht bekannt und 
auch nicht zu beziffern. 

Wie entwickelt sich die Weltbevölkerung? 

Wie viel CO2 stoßen wir in 30 Jahren aus?

Außerdem: Der Mensch wird durch Managemententscheidungen jene Bedingungen ständig ändern. Dennoch müssen wir heute 
schon Informationen bereitstellen und Abschätzungen liefern. Deshalb ist es so wichtig Wissenslücken zu
schließen und noch wichtiger, Unsicherheit durch Unwissen in Managaementprozesse einzubinden.

Mit dem **Klimamodell Explorer** können einige dieser Szenarien mit einander verglichen werden.
"""
)

_TRANSLATE_DE_INDICES = dict(
    title="Projektionen in einer Zahl darstellen",
    introduction="""Klimaprojektionen sind der Versuch möglichst viele verschiedene mögliche zukünftige
Szenarien mit in die Betrachtung des Klimawandels einzubeziehen. Durch verschiedene Unsicherheiten können
sich die einzelnen Modelle eines RCPs voneinander jedoch stärker unterscheiden, als zwei andere Modelle, die
sogar von unterschiedlichen Voraussetzungen ausgehen.
Hierdurch wird eine scharfe Aussage über die Folgen von Managementetscheidungen scheinbar unmöglich.
Vor allem, weil Trends in den Modellen so nur sehr schwer zu indentifizieren sind und entscheidende Änderungen
im Rauschen untergehen können.

Mit dem letzten Kapitel wird der Versuch unternommen, all die Variabilität und Unsicherheit auf eine Zahl
runterzubrechen, einem **Klimaindex**.
Dieser Index muss vor allem eine konkrete Bedeutung für einen Bestimmten **Kontext** haben, z.b. einem 
betriebswirtschaftlichen Entscheidungsprozess.

> *Soll ich in den nächsten Jahren eher Winterweizen oder Mais anbauen?*

Die Erkenntnis, dass alle Klimamodelle eine Erhöhung der Durchschnittstemperaturen in den nächsten Jahrzenten
vorhersagen ist für konkrete Entscheidungsprozesse nicht wichtig. Sondern, z.B. wenn Winterweizen weniger sensibel
auf Hitzetage reagiert, ist wichtig ob die Zahl dieser Tage erheblich zunimmt.

Für alle betrachteten Stationen in RUINS können im letzten Kapitel die Projektion der **Klimaindices** bis 2100 erforscht werden.
"""
)

_TRANSLATE_EN_INDICES = dict(
    title="Breaking down projections into one metric",
    introduction="""
"""
)

####
# OLD STUFF
def applySDM(wdata, data, meth='rel', cdf_threshold=0.9999999, lower_limit=0.1):
    '''apply structured distribution mapping to climate data and return unbiased version of dataset'''
    data_ub = data.copy()

    for k in data_ub.columns:
        data_col = data_ub[k].dropna()
        overlapx = pd.concat(
            [wdata.loc[data_col.index[0]:wdata.index[-1]], data_col.loc[data_col.index[0]:wdata.index[-1]]], axis=1)
        overlapx.columns = ['obs', 'cm']
        overlapx = overlapx.dropna()
        try:
            data_ub[k] = SDM(overlapx.obs, overlapx.cm, data_col, meth, cdf_threshold, lower_limit)
        except:
            data_ub[k] = data_ub[k] * np.nan

    data_ub[data_ub == 0.0000000] = np.nan
    data_ub = data_ub.loc[data_ub.index[0]:pd.to_datetime('2099-12-31 23:59:59')]

    return data_ub


def climate_indi(ts, indi='Summer days (Tmax ≥ 25°C)'):
    '''
    Calculate climate indicator days.
    Input time series of meteorological data
    '''
    if pd.infer_freq(ts.index) != 'D':
        print('Please provide daily data.')
        return

    # drop NA
    ts = ts.dropna()

    if indi == 'Summer days (Tmax ≥ 25°C)':  # summer days
        return (ts.Tmax >= 25.).groupby(ts.index.year).sum()
    elif indi == 'Ice days (Tmax < 0°C)':  # ice days
        return (ts.Tmax < 0.).groupby(ts.index.year).sum()
    elif indi == 'Frost days (Tmin < 0°C)':  # frost days
        return (ts.Tmin < 0.).groupby(ts.index.year).sum()
    elif indi == 'Hot days (Tmax ≥ 30°C)':  # hot days
        return (ts.Tmax >= 30.).groupby(ts.index.year).sum()
    elif indi == 'Tropic nights (Tmin ≥ 20°C)':  # tropic night
        return (ts.Tmin >= 20.).groupby(ts.index.year).sum()
    elif indi == 'Rainy days (Precip ≥ 1mm)':  # rainy days
        return (ts.Prec >= 1.).groupby(ts.index.year).sum()
    else:
        print('Nothing calculated.')
        return

# TODO: document + signature
# TODO: extract plotting
def climate_indices(dataManager: DataManager, config: Config):
    # get data
    weather = dataManager.read('weather')
    climate = dataManager.read('climate')

    # get the relevant settings
    stati = config.get('selected_station', 'coast')


    cindi = ['Ice days (Tmax < 0°C)', 'Frost days (Tmin < 0°C)', 'Summer days (Tmax ≥ 25°C)', 'Hot days (Tmax ≥ 30°C)','Tropic nights (Tmin ≥ 20°C)', 'Rainy days (Precip ≥ 1mm)']
    ci_topic = st.selectbox('Select Index:', cindi)

    if ci_topic == 'Rainy days (Precip ≥ 1mm)':
        vari = 'Prec'
        meth = 'rel'
    elif (ci_topic == 'Frost days (Tmin < 0°C)') | (ci_topic == 'Tropic nights (Tmin ≥ 20°C)'):
        vari = 'Tmin'
        meth = 'abs'
    else:
        vari = 'Tmax'
        meth = 'abs'

    w1 = weather[stati].sel(vars=vari).to_dataframe()
    w1.columns = ['bla', vari]

    fig = plt.figure(figsize=(10,2.5))
    wi = climate_indi(w1, ci_topic).astype(int)
    wi.plot(style='.', color='steelblue', label='Coast weather')
    wi.rolling(10, center=True).mean().plot(color='steelblue', label='Rolling mean\n(10 years)')

    if config['include_climate']:
        c1 = climate.sel(vars=vari).to_dataframe()
        c1 = c1[c1.columns[c1.columns != 'vars']]
        c2 = applySDM(w1[vari], c1, meth=meth)

        firstitem = True
        for j in c2.columns:
            dummyt = pd.DataFrame(c2[j])
            dummyt.columns = [vari]
            ci = pd.DataFrame(climate_indi(dummyt, ci_topic).astype(int))
            ci.columns = [j]

            if firstitem:
                ci[j].plot(style='.', color='gray', alpha=0.1, label='all climate projections')
                cid = ci
                firstitem = False
            else:
                ci[j].plot(style='.', color='gray', alpha=0.1, label='_')
                cid = pd.concat([cid, ci], axis=1)

        rcpx = np.array([x[-5:] for x in cid.columns.values])
        for n in np.unique(rcpx):
            nx = n
            if n == np.unique(rcpx)[-1]:
                nx = n + '\n(Rolling of 5 years)'
            cid[cid.columns[rcpx == n]].mean(axis=1).rolling(5, center=True).mean().plot(label='Mean of ' + nx)

    plt.legend(ncol=2)
    plt.ylabel('Number of days')
    plt.title(ci_topic)
    st.pyplot(fig)

    if ci_topic == 'Ice days (Tmax < 0°C)':
        st.markdown('''Number of days in one year which persistently remain below 0°C air temperature.''')
    elif ci_topic == 'Frost days (Tmin < 0°C)':
        st.markdown('''Number of days in one year which reached below 0°C air temperature.''')
    elif ci_topic == 'Summer days (Tmax ≥ 25°C)':
        st.markdown('''Number of days in one year which reached or exceeded 25°C air temperature.''')
    elif ci_topic == 'Hot days (Tmax ≥ 30°C)':
        st.markdown('''Number of days in one year which reached or exceeded 30°C air temperature.''')
    elif ci_topic == 'Tropic nights (Tmin ≥ 20°C)':
        st.markdown('''Number of days in one year which persistently remained above 20°C air temperature.''')
    elif ci_topic == 'Rainy days (Precip ≥ 1mm)':
        st.markdown('''Number of days in one year which received at least 1 mm precipitation.''')
    return


@partial_memoize(hash_names=['name', 'station', 'variable', 'time', '_filter'])
def _reduce_weather_data(dataManager: DataManager, name: str, variable: str, time: str, station: str = None, _filter: dict = None) -> pd.DataFrame:
    # get weather data
    arr: xr.Dataset = dataManager.read(name)

    if _filter is not None:
        arr = arr.filter_by_attrs(**_filter)

    if station is None:
        base = arr
    else:
        base = arr[station]

    # reduce to station and variable
    reduced = base.sel(vars=variable).resample(time=time)

    if variable == 'Tmax':
        df = reduced.max(dim='time').to_dataframe()
    elif variable == 'Tmin':
        df = reduced.min(dim='time').to_dataframe()
    else:
        df = reduced.mean(dim='time').to_dataframe()
    
    if station is None:
        return df.loc[:, df.columns != 'vars']
    else:
        return df[station]       


def climate_data_selector(dataManager: DataManager, config: Config, it: int = 0, variable: str = 'T', expander_container = st.sidebar, layout: str = 'columns', **kwargs):
    """Handles the selection and display of one paralell coordinates plot"""
    # create a unique key
    key = f'rcp_reference_{it}'
    title = f'Select Station #{it + 1}' if config.lang == 'en' else f'Station #{it + 1} auswählen'

    # get the container
    container = kwargs['container'] if 'container' in kwargs else st

    # if this is not the first iteration, we pre-select an item for rcp
    if it > 0:
        st.session_state[key] = 'rcp45'

    # make the data selection
    data_select.rcp_selector(
        dataManager,
        config,
        title=title,
        expander_container=expander_container,
        elements='__all__',
        layout=layout,
        RCP_KEY=key,
        allow_skip=False
    )

    # get the reference from config
    ref = config[key]

    # make the data sub-selection
    if ref == 'weather':
        data = dataManager.read('weather')
        drngx = (1980, 2000)
    else:
        data = dataManager.read('climate')
        drngx = (2050, 2070)

    # filter for rcps
    if ref in ('rcp26', 'rcp45', 'rcp85'):
        data = data.filter_by_attrs(RCP=ref)

    # create the data range slider
    drng = [pd.to_datetime(data.isel(time=0, vars=1).time.values).year, pd.to_datetime(data.isel(time=-1, vars=1).time.values).year]
    datarng = expander_container.slider('Date Range', drng[0], drng[1], drngx, key=f'dr{it}')

    # switch the variable
    if variable == 'T':
        afu = np.mean
    elif variable == 'Tmin':
        afu = np.min
    elif variable == 'Tmax':
        afu = np.max

    # aggregate
    dyp = data.sel(vars=variable).to_dataframe().resample('1M').apply(afu)
    dyp = dyp.loc[(dyp.index.year>=datarng[0]) & (dyp.index.year<datarng[1]),dyp.columns[dyp.columns!='vars']]

    # plot
    fig = climate_projection_parcoords(data=dyp, colorscale=kwargs.get('colorscale', 'electric'))
    container.plotly_chart(fig, use_container_width=True)

# TODO refactor the plots into the plotting module
def climate_plots(dataManager: DataManager, config: Config, expander_container = st.sidebar):
    """
    """
    cliproj = config['climate_scale']


    # TODO: build this into a Component ?
    if cliproj=='Regional':
        st.warning('Sorry, we currently have issues with the Regional model data. Please come back later.')
        st.stop()
        regaggs = ['North Sea Coast', 'Krummhörn',  'Niedersachsen', 'Inland']
        regagg = st.sidebar.selectbox('Spatial aggregation:', regaggs)

        if regagg=='North Sea Coast':
            climate = xr.load_dataset('data/cordex_coast.nc')
            climate.filter_by_attrs(RCP='rcp45')
        elif regagg=='Krummhörn':
            climate = xr.load_dataset('data/cordex_krummh.nc')

    # TODO: Refactor this part - similar stuff is used in weather explorer
    navi_vars = ['Maximum Air Temperature', 'Mean Air Temperature', 'Minimum Air Temperature']
    navi_var = expander_container.radio("Select variable:", options=navi_vars)
    if navi_var[:4] == 'Mini':
        vari = 'Tmin'
        ag = 'min'
    elif navi_var[:4] == 'Maxi':
        vari = 'Tmax'
        ag = 'max'
    else:
        vari = 'T'
        ag = 'mean'

    # add plots as needed
    num_of_plots = st.sidebar.number_input('# of datasets to compare', min_value=1, max_value=5, value=1)

    for it in range(int(num_of_plots)):
        # create expander
        plot_expander = st.expander(f'Temperatur (°C) Monthly {ag}', expanded=(it==num_of_plots - 1))
        left, right = plot_expander.columns((2, 8))
        left.markdown('### Options')
        opt = left.container()
        # add the colorbar as option
        cmap = left.selectbox(f'Plot #{it + 1} Colorbar', options=named_colorscales(), format_func=lambda l: l.capitalize())

        # add the Parcoords plot
        climate_data_selector(dataManager, config, it=it, variable=vari, colorscale=cmap, expander_container=opt, container=right)


def warming_data_plotter(dataManager: DataManager, config: Config):
    weather: xr.Dataset = dataManager.read('weather')
    statios = list(weather.keys())
    stat1 = config['selected_station']

    # build the placeholders
    plot_area = st.container()
    control_left, control_right = st.columns((1, 3))

    # TODO refactor in data-aggregator and data-plotter for different time frames


    # ----
    # data-aggregator controls
    navi_vars = ['Maximum Air Temperature', 'Mean Air Temperature', 'Minimum Air Temperature']
    navi_var = control_left.radio("Select variable:", options=navi_vars)
    if navi_var[:4] == 'Mini':
        vari = 'Tmin'
        ag = 'min'
    elif navi_var[:4] == 'Maxi':
        vari = 'Tmax'
        ag = 'max'
    else:
        vari = 'T'
        ag = 'mean'

    # controls end
    # ----

    # TODO: this produces a slider but also needs some data caching
    if config['temporal_agg'] == 'Annual':
        wdata = _reduce_weather_data(dataManager, name='weather', station=config['selected_station'], variable=vari, time='1Y')
        allw = _reduce_weather_data(dataManager, name='weather', variable=vari, time='1Y')

        dataLq = float(np.floor(allw.min().quantile(0.22)))
        datamin = float(np.min([dataLq, np.round(allw.min().min(), 1)]))
        
        if config['include_climate']:
            # get the rcp
            rcp = config['current_rcp']

            data = _reduce_weather_data(dataManager, name='cordex_coast', variable=vari, time='1Y', _filter=dict(RCP=rcp))
            data_ub = applySDM(wdata, data, meth='abs')

            dataUq = float(np.ceil(data_ub.max().quantile(0.76)))
            datamax = float(np.max([dataUq, np.round(data_ub.max().max(), 1)]))
        else:
            dataUq = float(np.ceil(allw.max().quantile(0.76)))
            datamax = float(np.max([dataUq,np.round(allw.max().max(), 1)]))

        datarng = control_right.slider('Adjust data range on x-axis of plot:', min_value=datamin, max_value=datamax, value=(dataLq, dataUq), step=0.1, key='drangew')

        # -------------------
        # start plotting plot
        if config['include_climate']:
            fig, ax = kde(wdata, data_ub.mean(axis=1), split_ts=3)
        else:
            fig, ax = kde(wdata, split_ts=3)

        ax.set_title(stat1 + ' Annual ' + navi_var)
        ax.set_xlabel('T (°C)')
        ax.set_xlim(datarng[0],datarng[1])
        plot_area.pyplot(fig)

        sndstat = st.checkbox('Show second station for comparison')

        if sndstat:
            stat2 = st.selectbox('Select second station:', [x for x in statios if x != config['selected_station']])
            wdata2 = _reduce_weather_data(dataManager, name='weather', station=stat2, variable=vari, time='1Y')

            fig, ax2 = kde(wdata2, split_ts=3)
            ax2.set_title(stat2 + ' Annual ' + navi_var)
            ax2.set_xlabel('T (°C)')
            ax2.set_xlim(datarng[0],datarng[1])
            st.pyplot(fig)

        # Re-implement this as a application wide service
        # expl_md = read_markdown_file('explainer/stripes.md')
        # st.markdown(expl_md, unsafe_allow_html=True)

    elif config['temporal_agg'] == 'Monthly':
        wdata = _reduce_weather_data(dataManager, name='weather', station=config['selected_station'], variable=vari, time='1M')

        ref_yr = control_right.slider('Reference period for anomaly calculation:', min_value=int(wdata.index.year.min()), max_value=2020,value=(max(1980, int(wdata.index.year.min())), 2000))

        if config['include_climate']:
            # get the rcp
            rcp = config['current_rcp']
            data = _reduce_weather_data(dataManager, name='cordex_coast', variable=vari, time='1M', _filter=dict(RCP=rcp))

            #ub = st.sidebar.checkbox('Apply SDM bias correction',True)
            ub = True # simplify here and automatically apply bias correction

            if ub:
                data_ub = applySDM(wdata, data, meth='abs')
                fig = yrplot_hm(pd.concat([wdata.loc[wdata.index[0]:data.index[0] - pd.Timedelta('1M')], data_ub.mean(axis=1)]), ref_yr, ag, li=2006, lang=config.lang)
            else:
                fig = yrplot_hm(pd.concat([wdata.loc[wdata.index[0]:data.index[0] - pd.Timedelta('1M')], data.mean(axis=1)]), ref_yr, ag, li=2006, lang=config.lang)

            # old matplotlib include
            #plt.title(stat1 + ' ' + navi_var + ' anomaly to ' + str(ref_yr[0]) + '-' + str(ref_yr[1]))
            #plot_area.pyplot(fig)
            fig.update_layout(title = f"{stat1} {navi_var} anomaly to {ref_yr[0]}-{ref_yr[1]}")
            plot_area.plotly_chart(fig, use_container_width=True)

            # compare to second station
            sndstat = st.checkbox('Compare to a second station?')
            
            if sndstat:
                stat2 = st.selectbox('Select second station:', [x for x in statios if x != stat1])
                wdata2 = _reduce_weather_data(dataManager, name='weather', station=stat2, variable=vari, time='1M')

                #ub = st.sidebar.checkbox('Apply SDM bias correction',True)
                ub = True # simplify here and automatically apply bias correction

                if ub:
                    data2_ub = applySDM(wdata2, data, meth='abs')
                    fig = yrplot_hm(pd.concat([wdata2.loc[wdata2.index[0]:data.index[0] - pd.Timedelta('1M')], data2_ub.mean(axis=1)]), ref_yr, ag, li=2006, lang=config.lang)
                else:
                    fig = yrplot_hm(pd.concat([wdata2.loc[wdata2.index[0]:data.index[0] - pd.Timedelta('1M')], data.mean(axis=1)]), ref_yr, ag, li=2006, lang=config.lang)
                
                fig.update_layout(title=f"{stat2} {navi_var} anomaly to {ref_yr[0]}-{ref_yr[1]}")
                plot_area.plotly_chart(fig, use_container_width=True)

        # TODO: break up this as well
        else:
            fig = yrplot_hm(sr=wdata, ref=ref_yr, ag=ag, lang=config.lang)
            fig.update_layout(title = f"{stat1} {navi_var} anomaly to {ref_yr[0]}-{ref_yr[1]}")
            plot_area.plotly_chart(fig, use_container_width=True)

            # old matplotlib include
            #plt.title(stat1 + ' ' + navi_var + ' anomaly to ' + str(ref_yr[0]) + '-' + str(ref_yr[1]))
            #plot_area.pyplot(fig)

            sndstat = st.checkbox('Compare to a second station?')

            if sndstat:
                stat2 = st.selectbox('Select second station:', [x for x in statios if x != stat1])
                data2 = _reduce_weather_data(dataManager, name='weather', station=stat2, variable=vari, time='1M')

                ref_yr2 = list(ref_yr)
                if ref_yr2[1]<data2.index.year.min():
                    ref_yr2[0] = data2.index.year.min()
                    ref_yr2[1] = ref_yr2[0]+10
                if ref_yr2[0]<data2.index.year.min():
                    ref_yr2[0] = data2.index.year.min()
                    if ref_yr2[1] - ref_yr2[0] < 10:
                        ref_yr2[1] = ref_yr2[0] + 10

                fig = yrplot_hm(sr=data2, ref=ref_yr2, ag=ag, lang=config.lang)
                fig.update_layout(title=f"{stat2} {navi_var} anomaly to {ref_yr2[0]}-{ref_yr2[1]}")
                plot_area.plotly_chart(fig, use_container_width=True)


def inject_cordex_overview(dataManager: DataManager, expanded: bool = False):
     with st.expander('CLIMATE MODEL OVERVIEW', expanded=expanded):
            # laod the cordex overview data
            overview = dataManager['cordex_overview'].read()

            # build the plot
            st.info('The Graph below groups all climate models available to RUINS into their global and regional family. Click on any element to expand it')
            fig = sunburst(overview, maxdepth=4)
            st.plotly_chart(fig, use_container_width=True)


def quick_access_buttons(config: Config, container = st.sidebar):
    """Add quick access button to skip parts of the Weather explorer"""
    # get the current stage
    stage = config.get('quick_access')

    # make columns
    l, r = container.columns(2)

    # make translations
    if config.lang == 'de':
        lab_weather = 'Wetterdaten Explorer'
        lab_climate = 'Klimamodell Explorer'
        lab_index = 'Klimadaten Indices'
    else:
        lab_weather = 'Weather explorer'
        lab_climate = 'Climate explorer'
        lab_index = 'Climate indices'

    # switch the cases
    if stage == 'climate':
        go_weather = l.button(lab_weather)
        go_climate = False
        go_idx = r.button(lab_index)
    elif stage == 'index':
        go_weather = l.button(lab_weather)
        go_climate = r.button(lab_climate)
        go_idx = False
    else:
        go_weather = False
        go_climate = l.button(lab_climate)
        go_idx = r.button(lab_index)

    # check if the Weather explorer is needed
    if go_weather:
        st.session_state.quick_access = 'weather'
        st.experimental_rerun()
    
    # check if the Climate explorer is needed
    if go_climate:
        if 'include_climate' in st.session_state:
            del st.session_state['include_climate']
        st.session_state.quick_access = 'climate'
        st.experimental_rerun()
    
    # check if the Climate indices are needed
    if go_idx:
        if 'include_climate' in st.session_state:
            del st.session_state['include_climate']
        st.session_state.quick_access = 'index'
        st.experimental_rerun()


def weather_stage(dataManager: DataManager, config: Config, data_expander=st.sidebar):
    # Story mode - go through each setting
    # update session state with current data settings
    data_expander = st.sidebar.expander('Data selection', expanded=True)
    data_select.data_select(dataManager, config, expander_container=data_expander, container=st)
    
    
    # build the app
    st.header('Weather Data Explorer')
    # TODO: move this text also into story mode?
    st.markdown('''In this section we provide visualisations to explore changes in observed weather data. Based on different variables and climate indices it is possible to investigate how climate change manifests itself in different variables, at different stations and with different temporal aggregation.''',unsafe_allow_html=True)

    warming_data_plotter(dataManager, config)

    # transition page
    st.markdown("""<hr style="margin-top: 4rem; margin-bottom: 2rem;" />""", unsafe_allow_html=True)
    st.success('Even about the present there is uncertainty! What about the future?')
    ok = st.button('LEARN MORE')

    if ok:
        st.session_state.quick_access = 'transition_climate'
        st.experimental_rerun()


def climate_stage(dataManager: DataManager, config: Config):
    # Story mode - go through each setting

    # update session state with current settings
    # get model scale
    option_container = st.sidebar.expander('OPTIONS', expanded=True)
    model_scale_select.model_scale_selector(dataManager, config, expander_container=option_container)

    # inject the overview
    inject_cordex_overview(dataManager)
    
    # run main visualization
    climate_plots(dataManager, config, expander_container=option_container)

    # transition page
    st.markdown("""<hr style="margin-top: 4rem; margin-bottom: 2rem;" />""", unsafe_allow_html=True)
    st.success('How do we make sense of this? Can we identify trends?')
    ok = st.button('LEARN MORE')

    if ok:
        st.session_state.quick_access = 'transition_index'
        st.experimental_rerun()


def indices_stage(dataManager: DataManager, config: Config, data_expander=st.sidebar):
    # Story mode - go through each setting
    # update session state with current data settings
    data_expander = st.sidebar.expander('Data selection', expanded=True)
    data_select.selected_station_selector(dataManager, config, expander_container=data_expander)
    data_select.rcp_selector(dataManager, config, expander_container=data_expander)

    # run actual visualization
    climate_indices(dataManager, config)


def transition_page(dataManager: DataManager, config: Config) -> None:
    """
    This Transition is shown when the user switches from weather explorer to
    climate projections or further to climate indices, without using the quick access buttons.
    The page can be used to present a primer how the two topics are related.
    """
    # check with transition page is needed
    if config['quick_access'] == 'transition_climate':
        t = config.translator(de=_TRANSLATE_DE_CLIMATE, en=_TRANSLATE_EN_CLIMATE)
        next_stage = 'climate'
    elif config['quick_access'] == 'transition_index':
        t = config.translator(de=_TRANSLATE_DE_INDICES, en=_TRANSLATE_EN_INDICES)
        next_stage = 'index'
    
    # build the page
    st.header(t('title'))
    st.markdown(t('introduction'), unsafe_allow_html=True)

    # add the sunburst plot
    if config['quick_access'] == 'transition_climate':
        inject_cordex_overview(dataManager, expanded=True)

    # add continue button
    ok = st.button('WEITER' if config.lang=='de' else 'CONTINUE')

    if ok:
        st.session_state.quick_access = next_stage
        st.experimental_rerun()
    else:
        st.stop()


def main_app(**kwargs):
    """Describe the params in kwargs here

    The main app has three 'stages': 
      
      * learning about weather
      * learning about climate
      * condensing info into climate indices

    
    """
    # build the config and dataManager from kwargs
    url_params = st.experimental_get_query_params()
    config, dataManager = build_config(url_params=url_params, **kwargs)

    # set page properties and debug view    
    st.set_page_config(page_title='Weather Explorer', layout=config.layout)
    debug_view.debug_view(dataManager, config, debug_name='DEBUG - initial state')

    # check if a stage was set
    if not config.has_key('quick_access'):
        st.session_state.quick_access = 'weather'
    stage = config['quick_access']

    # add the skip buttons
    btn_expander = st.sidebar.expander('QUICK ACCESS', expanded=True)
    quick_access_buttons(config, container=btn_expander)


    # -------------
    # Weather Stage
    if stage == 'weather':
        weather_stage(dataManager, config)
    elif stage == 'climate':
        climate_stage(dataManager, config)
    elif stage == 'index':
        indices_stage(dataManager, config)
    elif stage.startswith('transition'):
        transition_page(dataManager, config)
    else:
        st.error(f"We received weird data. A quick_access='{stage}' does not exist. Please contact the developer.")
        st.stop()
        
    # end state debug
    debug_view.debug_view(dataManager, config, debug_name='DEBUG - finished app')


if __name__ == '__main__':
    import fire
    fire.Fire(main_app)



import streamlit as st

from ruins.core import build_config, debug_view, Config
from ruins.plotting import plot_extreme_pdf

_TRANSLATE_DE = dict(
    title="Temperaturen verschieben sich",
    intro="""Dies ist ein langer introtext für dieses Beispiel""",
    temp_intro="""
Wähle zuerst die **mittlere Temperatur** von der du ausgehen möchtest.

Im zweiten Slider musst du die *Variabilität* der mittleren Temperatur angeben.
""",
    temp_outro="""Nun beschreibt die Abbildung den **Ist**-Zustand, also die Häufigkeit
mit der kalte oder warme Tage vorkommen und wie häufig es zu extremen Ereignissen kommt.
Im nächsten Schritt können wir die Erde erwärmen.
""",
    increase_intro="""In diesem Schritt kannst du nun die Durchschnittstemperatur auf
der Erde erhöhen. Eine Erhöhung um **1°C** Celsius mag nicht nach sehr viel klingen, und
natürlich sind die Schwankungen an einem einzigen Tag sehr viel höher, doch beobachte
genau, was mit den **dunkelroten Punkten** in der Abbildung passiert, wenn die Temperatur
sich erhöht.
""",
    increase_outro="""Wenn du die Abbildung genauer betrachtest siehst du zwei Dinge:

  * Es gibt deutlich mehr heiße Tage
  * Es gibt deutlich weniger kalte Tage

Mehr heiße Tage können wir gut mit Wetterbeobachtungen belegen. Allerdings kommt es 
nach wie vor zu **kalten** Extremereignissen. Dies legt nahe, dass sich auch die **Verteilung**
der durchschnittlichen Temperaturen geändert haben muss
""",
)

_TRANSLATE_EN = dict(
    title="Temperatures shift",
    intro="""This is a long intro-text for this example""",
    temp_intro="""
First select the **mean temperature** that you will rely on.

The second slider asks for the temperature's variability.
""",
    temp_outro="""The charts illustrates the current situation.
That is the frequency of warm and cold days as well as extremes.
As a next step, you can increase temperature on earth.
""",
    increase_intro="""Now it's time to increase the temperature on Earth.
A rise of **1°C** might not sound like much, and of course temperature is way more
volatile during the day, but watch out for the **darkred points** on the chart, when
you increase mean temperature.
""",
    increase_outro="""The chart suggest two main observations:

  * There are way more hot days
  * There are way less frost days

We find evidence for more hot days in our observation data. But other than the 
graph suggests, we still have cold and frost days. 
This implies that also the **distribution** of mean temperatures must have changed.
""",
)


def explainer(config: Config) -> None:
    """Introduce the concept"""
    # show the intro
    t = config.translator(de=_TRANSLATE_DE, en=_TRANSLATE_EN)
    st.title(t('title'))
    st.markdown(t('intro'))

    ok = st.button('GOT IT. Get started...')
    if ok:
        st.session_state.tshift_stage = 'temperature_intro'
        st.experimental_rerun()

    
def temperature_dist_intro(config: Config) -> None:
    """Introduce temperature distributions"""
    # explain
    t = config.translator(de=_TRANSLATE_DE, en=_TRANSLATE_EN)
    
    st.title(t('title'))
    plot = st.empty()
    st.markdown(t('temp_intro'))

    # build a slider
    mu = st.slider('Temperature', value=13.3, min_value=4., max_value=24., step=0.1, key='mu1')
    std = st.slider('Variability', value=0.1 * mu, min_value=0.0, max_value=10.0, step=0.01, key='std1')

    # make the plot
    fig = plot_extreme_pdf(mu, std)
    plot.plotly_chart(fig, use_container_width=True)

    # outro
    st.markdown("""<hr style="margin-top: 3rem; margin-bottom: 3rem;">""", unsafe_allow_html=True)
    st.markdown(t('temp_outro'))
    ok = st.button('ERDE ERWÄRMEN' if config.lang=='de' else 'WARM THE EARTH')
    if ok:
        st.session_state.tshift_stage = 't_increase'
        st.experimental_rerun()


def increase_temperatures(config: Config) -> None:
    """Increase the temperature on earth"""
    t = config.translator(de=_TRANSLATE_DE, en=_TRANSLATE_EN)

    # create the sidebar
    mu = st.sidebar.slider('Temperature', value=13.3, min_value=4., max_value=24., step=0.1, key='mu1')
    std = st.sidebar.slider('Variability', value=0.1 * mu, min_value=0.0, max_value=10.0, step=0.01, key='std1')

    # add the title
    st.title(t('title'))
    plot = st.empty()
    st.markdown(t('increase_intro'))

    # make the controller to increase temperature
    mu2 = st.slider('New temperature', min_value=mu, value=mu + 1.5, max_value=25., step=0.1, key='mu2')

    # make the plot
    fig = plot_extreme_pdf([mu, mu2], [std, std])
    plot.plotly_chart(fig, use_container_width=True)

    # outro
    st.markdown("""<hr style="margin-top: 3rem; margin-bottom: 3rem;">""", unsafe_allow_html=True)
    st.markdown(t('increase_outro'))
    
    # button
    ok = st.button('DIE VERTEILUNG ÄNDERN...' if config.lang=='de' else 'CHANGE THE DISTRIBUTION')
    if ok:
        st.session_state.tshift_stage = 'final'
        st.experimental_rerun()


def full_temperature_shift(config: Config) -> None:
    """Full application with all controls"""
    t = config.translator(de=_TRANSLATE_DE, en=_TRANSLATE_EN)

    # create sidebar
    st.sidebar.markdown('#### Present temperature')
    mu = st.sidebar.slider('Temperature', value=13.3, min_value=4., max_value=24., step=0.1, key='mu1')
    std = st.sidebar.slider('Variability', value=0.1 * mu, min_value=0.0, max_value=10.0, step=0.01, key='std1')
    
    st.sidebar.markdown('#### Future temperatures')
    mu2 = st.sidebar.slider('New temperature', min_value=mu, value=mu + 1.5, max_value=25., step=0.1, key='mu2')
    std2 = st.sidebar.slider('Variability', value=0.1 * mu2, min_value=0.0, max_value=10.0, step=0.01, key='std2')

    # add the title
    st.title(t('title'))
    
    # make the plot
    fig = plot_extreme_pdf([mu, mu2], [std, std2])
    fig = st.plotly_chart(fig, use_container_width=True)


def main_app(**kwargs):
    # build the config and dataManager from kwargs
    url_params = st.experimental_get_query_params()
    config, _ = build_config(url_params=url_params, omit_dataManager=True, **kwargs)

    # set page properties and debug view
    st.set_page_config(page_title='Temperature shift', layout=config.layout)
    debug_view.debug_view(None, config, 'DEBUG - initial state')

    # MAIN APP
    # -----------
    stage = config.get('tshift_stage', 'intro')
    if stage == 'intro':
        explainer(config)
    elif stage == 'temperature_intro':
        temperature_dist_intro(config)
    elif stage == 't_increase':
        increase_temperatures(config)
    elif stage == 'final':
        full_temperature_shift(config)
    else:
        st.error("""The application is in an undefinded state. Please contact the developer with the info below""")
        st.json({'message': 'tshift_stage is unkown', 'value': stage, 'origin': 'temperature_shift.py',})
        st.stop()


    # end state debug
    debug_view.debug_view(None, config, 'DEBUG - final state')



if __name__ == '__main__':
    import fire
    fire.Fire(main_app)



import streamlit as st

from ruins.core import build_config, debug_view, DataManager, Config


_TRANSLATE_EN = dict(
    title='Extreme events & flooding',
    introduction="""
    In this section, a model is presented to assess the influence of sea level change, inland precipitation, and 
    management options on flooding risks in a below sea level and thus drained area in northern Germany.
"""
)

_TRANSLATE_DE = dict(
    title='Extremereignisse & Überflutungen',
    introduction="""
    In diesem Abschnitt wird ein Modell vorgestellt, mit dem sich der Einfluss von Meeresspiegelveränderungen, 
    Inlandsniederschlägen und Managementoptionen auf Überflutungsrisiken in einem unterhalb des Meeresspiegels 
    liegenden und damit entwässerten Gebiet in Norddeutschland auswirken.
"""
)


def concept_explainer(config: Config, **kwargs):
    """Show an explanation, if it was not already shown.
    """
    # check if we saw the explainer already
    if config.has_key('extremes_explainer'):
        return
    
    # get the container and a translation function
    container = kwargs['container'] if 'container' in kwargs else st
    t = config.translator(en=_TRANSLATE_EN, de=_TRANSLATE_DE)

    # place title and intro
    container.title(t('title'))
    container.markdown(t('introduction'), unsafe_allow_html=True)

    # check if the user wants to continue
    accept = container.button('WEITER' if config.lang == 'de' else 'CONTINUE')
    if accept:
        st.session_state.extremes_explainer = True
        st.experimental_rerun()
    else:
        st.stop()


def demonstrate_flood_model(config:Config, **kwargs):
    """
    Demonstrate the flooding model using a real-life flood disaster.
    """
    container = kwargs['container'] if 'container' in kwargs else st
    t = config.translator(de=_TRANSLATE_DE, en=_TRANSLATE_EN)

    st.empty()


def flood_model(config:Config, **kwargs):
    """
    Version of the flooding model in which the user can play around with the parameters.
    """
    container = kwargs['container'] if 'container' in kwargs else st
    t = config.translator(de=_TRANSLATE_DE, en=_TRANSLATE_EN)

    st.empty()
    

def main_app(**kwargs):
    """
    """
    # build the config and dataManager from kwargs
    url_params = st.experimental_get_query_params()
    config, dataManager = build_config(url_params=url_params, **kwargs)

    # set page properties and debug view    
    st.set_page_config(page_title='Sea level rise Explorer', layout=config.layout)
    debug_view.debug_view(dataManager, config, debug_name='DEBUG - initial state')

    # explainer
    concept_explainer(config)

    # TODO: Demonstrate model using one or two real flood events.
    demonstrate_flood_model(config)

    # TODO: expert mode: user takes control over model parameters
    flood_model(config)

    # end state debug
    debug_view.debug_view(dataManager, config, debug_name='DEBUG - finished app')


if __name__ == '__main__':
    import fire
    fire.Fire(main_app)



from typing import Callable, List
from functools import wraps
import hashlib


LOCAL = dict()

def _hashargs(fname, argnames):
    h = f'{fname};' + ','.join(argnames)
    digest = hashlib.sha256(h.encode()).hexdigest()

    return digest

def partial_memoize(hash_names: List[str], store: str = 'local'):
    def func_decorator(f: Callable):
        @wraps(f)
        def wrapper(*args, **kwargs):
            argnames = [str(a) for a in args if a in hash_names]
            argnames.extend([str(v) for k, v in kwargs.items() if k in hash_names])
            
            # get the parameter hash
            h = _hashargs(f.__name__, argnames)
            
            # check if result exists
            if h in LOCAL:
                return LOCAL.get(h)
            else:
                # process
                result = f(*args, **kwargs)
                LOCAL[h] = result
                return result
        return wrapper
    return func_decorator
